# Python functions for NucDynamics
import sys
import numpy as np
import multiprocessing
from multiprocessing import Queue, Process
from time import time
import os
import gzip
from sc3dg.model import STARK_dyn_3dg as dyn_3dg

PROG_NAME = 'nuc_dynamics'
DESCRIPTION = 'Single-cell Hi-C genome and chromosome structure calculation module for 3DG'

N3D = 'n3d'
PDB = 'pdb'
FORMATS = [N3D, PDB]
MAX_CORES = multiprocessing.cpu_count()

def warn(msg, prefix='WARNING'):

  print('%8s : %s' % (prefix, msg))


def critical(msg, prefix='ABORT'):

  print('%8s : %s' % (prefix, msg))
  sys.exit(0)
  
  
def _parallel_func_wrapper(queue, target_func, proc_data, common_args):
  
  for t, data_item in proc_data:
    result = target_func(data_item, *common_args)
    
    if queue:
      queue.put((t, result))
    
    elif isinstance(result, Exception):
      raise(result)


def parallel_split_job(target_func, split_data, common_args, num_cpu=MAX_CORES, collect_output=True):
  num_tasks = len(split_data)
  num_process = min(num_cpu, num_tasks)

  processes = []
  queue = Queue() if collect_output else None

  for p in range(num_process):
    proc_data = [(t, data_item) for t, data_item in enumerate(split_data) if t % num_process == p]
    args = (queue, target_func, proc_data, common_args)
    proc = Process(target=_parallel_func_wrapper, args=args)
    processes.append(proc)

  for proc in processes:
    proc.start()

  if collect_output:
    results = [None] * num_tasks
    for _ in range(num_tasks):
      t, result = queue.get()
      if isinstance(result, Exception):
        print('\n* * * * C/Cython code may need to be recompiled. Try running "python setup_cython.py build_ext --inplace" * * * *\n')
        raise result
      results[t] = result
    queue.close()
    queue.join_thread()
    return results
  else:
    for proc in processes:
      proc.join()


def load_pair_file(file_path, target_Chrom):
  """Load chromosome and contact data from NCC format file, as output from NucProcess"""
  
  if file_path.endswith('.gz'):
    import gzip
    file_obj = gzip.open(file_path,'rb')
  
  else:
    file_obj = open(file_path) 
  
  # Observations are treated individually in single-cell Hi-C,
  # i.e. no binning, so num_obs always 1 for each contact
  num_obs = 1  
    
  contact_dict = {}
  chromosomes = set()
    
  for line in file_obj:
    line = line.decode()
    if "#" == line[0]:
      continue
    #chr_a, f_start_a, f_end_a, start_a, end_a, strand_a, chr_b, f_start_b, f_end_b, start_b, end_b, strand_b, ambig_group, pair_id, swap_pair = line.split()
    # print(line.split('\t'))
    PairId, chr_a, f_start_a, chr_b, f_start_b, strand_a, strand_b, pairType, mapq1, mapq2 = line.split("\t")[0:10]

    if chr_a =="!" or chr_b =="!" or pairType != "UU":
      continue

    if chr_a not in target_Chrom or chr_b not in target_Chrom:
      continue

    pos_a = int(f_start_a)
    pos_b = int(f_start_b)
 
    if chr_a > chr_b:
      chr_a, chr_b = chr_b, chr_a
      pos_a, pos_b = pos_b, pos_a
    
    if chr_a not in contact_dict:
      contact_dict[chr_a] = {}
      chromosomes.add(chr_a)
      
    if chr_b not in contact_dict[chr_a]:
      contact_dict[chr_a][chr_b] = [] 
      chromosomes.add(chr_b)
        
    contact_dict[chr_a][chr_b].append((pos_a, pos_b, num_obs, int(1)))
   
  file_obj.close()

  contact_dict = {key: contact_dict[key] for key in sorted(contact_dict.keys())}
  chromo_limits = {}
    
  for chr_a in contact_dict:
    for chr_b in contact_dict[chr_a]:
      contacts = np.array(contact_dict[chr_a][chr_b]).T
      contact_dict[chr_a][chr_b] = contacts
      
      seq_pos_a = contacts[0]
      seq_pos_b = contacts[1]
      
      min_a = min(seq_pos_a)
      max_a = max(seq_pos_a)
      min_b = min(seq_pos_b)
      max_b = max(seq_pos_b)
        
      if chr_a in chromo_limits:
        prev_min, prev_max = chromo_limits[chr_a]
        chromo_limits[chr_a] = [min(prev_min, min_a), max(prev_max, max_a)]
      else:
        chromo_limits[chr_a] = [min_a, max_a]
      
      if chr_b in chromo_limits:
        prev_min, prev_max = chromo_limits[chr_b]
        chromo_limits[chr_b] = [min(prev_min, min_b), max(prev_max, max_b)]
      else:
        chromo_limits[chr_b] = [min_b, max_b]
         
  chromosomes = sorted(chromosomes)      
        
  return chromosomes, chromo_limits, contact_dict


def export_n3d_coords(file_path, coords_dict, seq_pos_dict):
  
  file_obj = open(file_path, 'w')
  write = file_obj.write
  
  for chromo in seq_pos_dict:
    chromo_coords = coords_dict[chromo]
    chromo_seq_pos = seq_pos_dict[chromo]
    
    num_models = len(chromo_coords)
    num_coords = len(chromo_seq_pos)
    
    line = '%s\t%d\t%d\n' % (chromo, num_coords, num_models)
    write(line)
    
    for j in range(num_coords):
      data = chromo_coords[:,j].ravel().tolist()
      data = '\t'.join('%.8f' % d for d in  data)
      
      line = '%d\t%s\n' % (chromo_seq_pos[j], data)
      write(line)

  file_obj.close()


def export_pdb_coords(file_path, coords_dict, seq_pos_dict, particle_size, scale=1.0, extended=True):
  """
  Write chromosome particle coordinates as a PDB format file
  """

  alc = ' '
  ins = ' '
  prefix = 'HETATM'
  line_format = '%-80.80s\n'
  
  if extended:
    pdb_format = '%-6.6s%5.1d %4.4s%s%3.3s %s%4.1d%s   %8.3f%8.3f%8.3f%6.2f%6.2f          %2.2s  %10d\n'
    ter_format = '%-6.6s%5.1d      %s %s%4.1d%s                                                     %10d\n'
  else:
    pdb_format = '%-6.6s%5.1d %4.4s%s%3.3s %s%4.1d%s   %8.3f%8.3f%8.3f%6.2f%6.2f          %2.2s  \n'
    ter_format = '%-6.6s%5.1d      %s %s%4.1d%s                                                     \n'

  file_obj = open(file_path, 'w')
  write = file_obj.write
  
  chromosomes = list(seq_pos_dict.keys())
  
  sort_chromos = []
  for chromo in chromosomes:
    if chromo[:3] == 'chr':
      key = chromo[3:]
    else:
      key = chromo
    
    if key.isdigit():
      key = '%03d' % int(key)
    
    sort_chromos.append((key, chromo))
  
  sort_chromos.sort()
  sort_chromos = [x[1] for x in sort_chromos]
  
  num_models = len(coords_dict[chromosomes[0]])
  title = 'NucDynamics genome structure export'
  
  write(line_format % 'TITLE     %s' % title)
  write(line_format % 'REMARK 210') 
  write(line_format % 'REMARK 210 Atom type C is used for all particles')
  write(line_format % 'REMARK 210 Atom number increases every %s bases' % particle_size)
  write(line_format % 'REMARK 210 Residue code indicates chromosome')
  write(line_format % 'REMARK 210 Residue number represents which sequence Mb the atom is in')
  write(line_format % 'REMARK 210 Chain letter is different every chromosome, where Chr1=a, Chr2=b etc.')
  
  if extended:
    file_obj.write(line_format % 'REMARK 210 Extended PDB format with particle seq. pos. in last column')
  
  file_obj.write(line_format % 'REMARK 210')
  
  pos_chromo = {}
  
  for m in range(num_models):
    line = 'MODEL     %4d' % (m+1)
    write(line_format  % line)
    
    c = 0
    j = 1
    seqPrev = None
    
    for k, chromo in enumerate(sort_chromos):
      chain_code = chr(ord('a')+k)            
      
      tlc = chromo
      while len(tlc) < 2:
        tlc = '_' + tlc
      
      if len(tlc) == 2:
        tlc = 'C' + tlc
      
      if len(tlc) > 3:
        tlc = tlc[:3]
      
      chromo_model_coords = coords_dict[chromo][m]
      
      if not len(chromo_model_coords):
        continue
      
      pos = seq_pos_dict[chromo]
      
      for i, seqPos in enumerate(pos):
        c += 1
 
        seqMb = int(seqPos//1e6) + 1
        
        if seqMb == seqPrev:
          j += 1
        else:
          j = 1
        
        el = 'C'
        a = 'C%d' % j
          
        aName = '%-3s' % a
        x, y, z = chromo_model_coords[i] #XYZ coordinates
         
        seqPrev = seqMb
        pos_chromo[c] = chromo
        
        if extended:
          line  = pdb_format % (prefix,c,aName,alc,tlc,chain_code,seqMb,ins,x,y,z,0.0,0.0,el,seqPos)
        else:
          line  = pdb_format % (prefix,c,aName,alc,tlc,chain_code,seqMb,ins,x,y,z,0.0,0.0,el)
          
        write(line)
 
    write(line_format  % 'ENDMDL')
 
  for i in range(c-2):
     if pos_chromo[i+1] == pos_chromo[i+2]:
       line = 'CONECT%5.1d%5.1d' % (i+1, i+2)
       write(line_format  % line)
 
  write(line_format  % 'END')
  file_obj.close()


def remove_isolated_contacts(contact_dict, threshold=int(2e6)):
  """
  Select only contacts which are within a given sequence separation of another
  contact, for the same chromosome pair
  """
    
  for chromoA in contact_dict:
    print(chromoA)
    for chromoB in contact_dict[chromoA]:
      contacts = contact_dict[chromoA][chromoB]
      positions = np.array(contacts[:2], np.int32).T
      
      if len(positions): # Sometimes empty e.g. for MT, Y chromos 
        active_idx = dyn_3dg.getSupportedPairs(positions, np.int32(threshold))
        contact_dict[chromoA][chromoB] = contacts[:,active_idx]
     
  return contact_dict
  

def remove_violated_contacts(contact_dict, coords_dict, particle_seq_pos, particle_size, threshold=5.0):  
  """
  Remove contacts whith structure distances that exceed a given threshold
  """
  
  for chr_a in contact_dict:
    for chr_b in contact_dict[chr_a]:
      contacts = contact_dict[chr_a][chr_b]
      
      contact_pos_a = contacts[0].astype(np.int32)
      contact_pos_b = contacts[1].astype(np.int32)
      
      coords_a = coords_dict[chr_a]
      coords_b = coords_dict[chr_b]

      struc_dists = []
 
      for m in range(len(coords_a)):
        coord_data_a = dyn_3dg.getInterpolatedCoords([chr_a], {chr_a:contact_pos_a}, particle_seq_pos, coords_a[m])
        coord_data_b = dyn_3dg.getInterpolatedCoords([chr_b], {chr_b:contact_pos_b}, particle_seq_pos, coords_b[m])
 
        deltas = coord_data_a - coord_data_b
        dists = np.sqrt((deltas*deltas).sum(axis=1))
        struc_dists.append(dists)
      
      # Average over all conformational models
      struc_dists = np.array(struc_dists).T.mean(axis=1)
      
      # Select contacts with distances below distance threshold
      indices = (struc_dists < threshold).nonzero()[0]
      contact_dict[chr_a][chr_b] = contacts[:,indices]
        
  return contact_dict
  
  
def get_random_coords(pos_dict, chromosomes, num_models, radius=10.0):
  """
  Get random, uniformly sampled coorinate positions, restricted to
  a sphere of given radius
  """
    
  from numpy.random import uniform
    
  num_particles = sum([len(pos_dict[chromo]) for chromo in chromosomes])
  coords = np.empty((num_models, num_particles, 3))
  r2 = radius*radius
    
  for m in range(num_models):
    
    for i in range(num_particles):
      x = y = z = radius

      while x*x + y*y + z*z >= r2:
        x = radius * (2*uniform(0,1) - 1)
        y = radius * (2*uniform(0,1) - 1)
        z = radius * (2*uniform(0,1) - 1)
      
      coords[m,i] = [x,y,z]

  return coords
  
  
def pack_chromo_coords(coords_dict, chromosomes):
  """
  Place chromosome 3D coordinates stored in a dictionary keyed by
  chromosome name into a single, ordered array. The chromosomes argument
  is required to set the correct array storage order.
  """

  chromo_num_particles = [len(coords_dict[chromo][0]) for chromo in chromosomes]
  n_particles = sum(chromo_num_particles)
  n_models = len(coords_dict[chromosomes[0]])  
  coords = np.empty((n_models, n_particles, 3), float)
  
  j = 0
  for i, chromo in enumerate(chromosomes):
    span = chromo_num_particles[i]
    coords[:,j:j+span] = coords_dict[chromo]
    j += span
      
  return coords
  
 
def unpack_chromo_coords(coords, chromosomes, seq_pos_dict):
  """
  Exctract coords for multiple chromosomes stored in a single array into
  a dictionary, keyed by chromosome name. The chromosomes argument is required
  to get the correct array storage order.
  """

  chromo_num_particles = [len(seq_pos_dict[chromo]) for chromo in chromosomes]
  n_seq_pos = sum(chromo_num_particles)
  n_models, n_particles, dims = coords.shape

  if n_seq_pos != n_particles:
    msg = 'Model coordinates must be an array of num models x %d' % (n_seq_pos,)
    raise(Exception(msg))  
  
  coords_dict = {}
        
  j = 0
  for i, chromo in enumerate(chromosomes):
    span = chromo_num_particles[i]
    coords_dict[chromo] = coords[:,j:j+span] # all models, slice
    j += span
 
  return coords_dict


def anneal_model(model_data, anneal_schedule, masses, radii, restraint_indices, restraint_dists,
                 ambiguity, temp, time_step, dyn_steps, repulse, n_rep_max):

  import gc
  
  m, model_coords = model_data
  
  # Anneal one model in parallel
  
  time_taken = 0.0
  
  if m == 0:
    printInterval = max(1, dyn_steps/2)
  
  else:
    printInterval = 0
  
  print('  starting model %d' % m)
  
  for temp, repulse in anneal_schedule:
    gc.collect() # Try to free some memory
    
    # Update coordinates for this temp
    
    try:  
      dt, n_rep_max = dyn_3dg.runDynamics(model_coords, masses, radii, restraint_indices, restraint_dists,
                                           ambiguity, temp, time_step, dyn_steps, repulse, nRepMax=n_rep_max,
                                           printInterval=printInterval)
    
    except Exception as err:
      return err
 
    n_rep_max = np.int32(1.05 * n_rep_max) # Base on num in prev cycle, plus a small overhead
    time_taken += dt
  
  # Center
  model_coords -= model_coords.mean(axis=0)

  print('  done model %d' % m)
  
  return model_coords
 
  
def anneal_genome(chromosomes, contact_dict, num_models, particle_size,
                  general_calc_params, anneal_params,
                  prev_seq_pos_dict=None, start_coords=None, num_cpu=MAX_CORES):
    """
    Use chromosome contact data to generate distance restraints and then
    apply a simulated annealing protocul to generate/refine coordinates.
    Starting coordinates may be random of from a previous (e.g. lower
    resolution) stage.
    """
    
    from numpy import random
    from math import log, exp, atan, pi
    
    random.seed(general_calc_params['random_seed'])
    particle_size = np.int32(particle_size)
    
    # Calculate distance restrains from contact data   
    restraint_dict, seq_pos_dict = dyn_3dg.calc_restraints(chromosomes, contact_dict, particle_size,
                                                   scale=1.0, exponent=general_calc_params['dist_power_law'],
                                                   lower=general_calc_params['contact_dist_lower'],
                                                   upper=general_calc_params['contact_dist_upper'])
    
    # Concatenate chromosomal data into a single array of particle restraints
    # for structure calculation. Add backbone restraints between seq. adjasent particles.
    restraint_indices, restraint_dists = dyn_3dg.concatenate_restraints(restraint_dict, seq_pos_dict, particle_size,
                                                                general_calc_params['backbone_dist_lower'],
                                                                general_calc_params['backbone_dist_upper'])
 
    # Setup starting structure
    if (start_coords is None) or (prev_seq_pos_dict is None):
      coords = get_random_coords(seq_pos_dict, chromosomes, num_models,
                                 general_calc_params['random_radius'])
      
      num_coords = coords.shape[1]
            
    else:
      # Convert starting coord dict into single array
      coords = pack_chromo_coords(start_coords, chromosomes)
      num_coords = sum([len(seq_pos_dict[c]) for c in chromosomes])
        
      if coords.shape[1] != num_coords: # Change of particle_size
        interp_coords = np.empty((num_models, num_coords, 3))
        
        for m in range(num_models): # Starting coords interpolated from previous particle positions
          interp_coords[m] = dyn_3dg.getInterpolatedCoords(chromosomes, seq_pos_dict, prev_seq_pos_dict, coords[m])
        
        coords = interp_coords
        
    # Equal unit masses and radii for all particles
    masses = np.ones(num_coords,  float)
    radii = np.ones(num_coords,  float)
    
    # Ambiguiity strides not used here, so set to 1
    num_restraints = len(restraint_indices)
    ambiguity = np.ones(num_restraints,  np.int32)
        
    # Below will be set to restrict memory allocation in the repusion list
    # (otherwise all vs all can be huge)
    n_rep_max = np.int32(0)
    
    # Annealing parameters
    temp_start = anneal_params['temp_start']
    temp_end = anneal_params['temp_end']
    temp_steps = anneal_params['temp_steps']
    
    # Setup annealig schedule: setup temps and repulsive terms
    adj = 1.0 / atan(10.0)
    decay = log(temp_start/temp_end)    
    anneal_schedule = []
    
    for step in range(temp_steps):
      frac = step/float(temp_steps)
    
      # exponential temp decay
      temp = temp_start * exp(-decay*frac)
    
      # sigmoidal repusion scheme
      repulse = 0.5 + adj * atan(frac*20.0-10) / pi 
      
      anneal_schedule.append((temp, repulse))  
        
    # Paricle dynamics parameters
    # (these need not be fixed for all stages, but are for simplicity)    
    dyn_steps = anneal_params['dynamics_steps']
    time_step = anneal_params['time_step']
       
    # Update coordinates in the annealing schedule which is applied to each model in parallel
    common_args = [anneal_schedule, masses, radii, restraint_indices, restraint_dists,
                   ambiguity, temp, time_step, dyn_steps, repulse, n_rep_max]
    
    task_data = [(m, coords[m]) for m in range(len(coords))]

    coords = parallel_split_job(anneal_model, task_data, common_args, int(num_cpu), collect_output=True)
    coords = np.array(coords)
   
    # Convert from single coord array to dict keyed by chromosome
    coords_dict = unpack_chromo_coords(coords, chromosomes, seq_pos_dict)
    
    return coords_dict, seq_pos_dict


def open_file(file_path, mode=None, gzip_exts=('.gz','.gzip')):
  """
  GZIP agnostic file opening
  """
  IO_BUFFER = int(4e6)
  
  if os.path.splitext(file_path)[1].lower() in gzip_exts:
    file_obj = gzip.open(file_path, mode or 'rt')
  else:
    file_obj = open(file_path, mode or 'rU', IO_BUFFER)
 
  return file_obj

def load_n3d_coords(file_path):
 """
 Load genome structure coordinates and particle sequence positions from an N3D format file.

 Args:
     file_path: str ; Location of N3D (text) format file

 Returns:
     dict {str:ndarray(n_coords, int)}                  ; {chromo: seq_pos_array}
     dict {str:ndarray((n_models, n_coords, 3), float)} ; {chromo: coord_3d_array}

 """
 seq_pos_dict = {}
 coords_dict = {}

 with open_file(file_path) as file_obj:
   chromo = None

   for line in file_obj:

     data = line.split()
     n_items = len(data)

     if not n_items:
       continue

     elif data[0] == '#':
       continue

     elif n_items == 3:
       chromo, n_coords, n_models = data

       #if chromo.lower()[:3] == 'chr':
       #  chromo = chromo[3:]

       n_coords = int(n_coords)
       n_models = int(n_models)

       #chromo_seq_pos = np.empty(n_coords, int)
       chromo_seq_pos = np.empty(n_coords, 'int32')
       chromo_coords = np.empty((n_models, n_coords, 3), float)

       coords_dict[chromo]  = chromo_coords
       seq_pos_dict[chromo] = chromo_seq_pos

       check = (n_models * 3) + 1
       i = 0

     elif not chromo:
       raise Exception('Missing chromosome record in file %s' % file_path)

     elif n_items != check:
       msg = 'Data size in file %s does not match Position + Models * Positions * 3'
       raise Exception(msg % file_path)

     else:
       chromo_seq_pos[i] = int(data[0])

       coord = [float(x) for x in data[1:]]
       coord = np.array(coord).reshape(n_models, 3)
       chromo_coords[:,i] = coord
       i += 1

 return seq_pos_dict, coords_dict


def export_coords(out_format, out_file_path, coords_dict, particle_seq_pos, particle_size):
  
  # Save final coords as N3D or PDB format file
  
  if out_format == PDB:
    if not out_file_path.endswith(PDB):
      out_file_path = '%s.%s' % (out_file_path, PDB)
  
    export_pdb_coords(out_file_path, coords_dict, particle_seq_pos, particle_size)
 
  else:
    if not out_file_path.endswith(N3D):
      out_file_path = '%s.%s' % (out_file_path, N3D)
      
    export_n3d_coords(out_file_path, coords_dict, particle_seq_pos)
    
  print('Saved structure file to: %s' % out_file_path)


def calc_genome_structure(ncc_file_path, out_file_path, general_calc_params, anneal_params,
                          particle_sizes, num_models=5, isolation_threshold=2e6,
                          out_format=N3D, num_cpu=MAX_CORES,Target_Chrom=None,
                          start_coords_path=None, save_intermediate=False):

  from time import time

  # Load single-cell Hi-C data from NCC contact file, as output from NucProcess
  chromosomes, chromo_limits, contact_dict = load_pair_file(ncc_file_path,target_Chrom=Target_Chrom)

  # Only use contacts which are supported by others nearby in sequence, in the initial instance
  contact_dict = remove_isolated_contacts(contact_dict, threshold=isolation_threshold)

  # Initial coords will be random
  start_coords = None

  # Record particle positions from previous stages
  # so that coordinates can be interpolated to higher resolution
  prev_seq_pos = None

  if start_coords_path:
    prev_seq_pos, start_coords = load_n3d_coords(start_coords_path)
    if start_coords:
      chromo = next(iter(start_coords)) # picks out arbitrary chromosome
      num_models = len(start_coords[chromo])
    
  for stage, particle_size in enumerate(particle_sizes):
 
      print("Running structure caculation stage %d (%d kb)" % (stage+1, (particle_size/1e3)))
 
      # Can remove large violations (noise contacts inconsistent with structure)
      # once we have a reasonable resolution structure
      
      if stage > 0:
        if particle_size < 0.5e6:
            remove_violated_contacts(contact_dict, coords_dict, particle_seq_pos,
                                     particle_size, threshold=6.0)
        elif particle_size < 0.25e6:
            remove_violated_contacts(contact_dict, coords_dict, particle_seq_pos,
                                     particle_size, threshold=5.0)
 
      coords_dict, particle_seq_pos = anneal_genome(chromosomes, contact_dict, num_models, particle_size,
                                                    general_calc_params, anneal_params,
                                                    prev_seq_pos, start_coords, num_cpu)
 
      if save_intermediate and stage < len(particle_sizes)-1:
        file_path = '%s_%d.%s' % (out_file_path[:-4], stage, out_file_path[-3:]) # DANGER: assumes that suffix is 3 chars
        export_coords(out_format, file_path, coords_dict, particle_seq_pos, particle_size)
        
      # Next stage based on previous stage's 3D coords
      # and their respective seq. positions
      start_coords = coords_dict
      prev_seq_pos = particle_seq_pos

  # Save final coords
  export_coords(out_format, out_file_path, coords_dict, particle_seq_pos, particle_size)




import numpy as np
import random
from math import sqrt
import math
#from libc.math cimport  ceil

BOLTZMANN_K = 0.0019872041

class NucCythonError(Exception):
    def __init__(self, err):
        super().__init__(err)

def getRepulsionList(rep_list, coords, regions_1, regions_2, idx_1, idx_2, s1, s2, nCoords, nRepMax, rep_dist, radii, max_radius):

    n = 0  # Num close pairs
    s3 = int(s2 / s1)
    n_rep_found = 0

    rep_dist2 = rep_dist + max_radius

    if nCoords < s2:  # No split
        for i in range(nCoords - 2):
            for j in range(i + 2, nCoords):
                d_lim = rep_dist + radii[i] + radii[j]

                dx = coords[i, 0] - coords[j, 0]
                if abs(dx) > d_lim:
                    continue

                dy = coords[i, 1] - coords[j, 1]
                if abs(dy) > d_lim:
                    continue

                dz = coords[i, 2] - coords[j, 2]
                if abs(dz) > d_lim:
                    continue

                d2 = dx * dx + dy * dy + dz * dz

                d_lim2 = d_lim * d_lim
                if d2 > d_lim2:
                    continue

                if n < nRepMax:
                    rep_list[n, 0] = i
                    rep_list[n, 1] = j
                    n += 1

                n_rep_found += 1


    elif nCoords < s1 * s2:  # Single split
        # Calc bounding X regions
        n1 = 0
        for i in range(0, nCoords, s1):

            for k in range(3):
                regions_1[n1, k, 0] = coords[i, k]
                regions_1[n1, k, 1] = coords[i, k]

            for j in range(i + 1, i + s1):
                if j >= nCoords:
                    break

                if coords[j, 0] < regions_1[n1, 0, 0]:
                    regions_1[n1, 0, 0] = coords[j, 0]

                if coords[j, 0] > regions_1[n1, 0, 1]:
                    regions_1[n1, 0, 1] = coords[j, 0]

                if coords[j, 1] < regions_1[n1, 1, 0]:
                    regions_1[n1, 1, 0] = coords[j, 1]

                if coords[j, 1] > regions_1[n1, 1, 1]:
                    regions_1[n1, 1, 1] = coords[j, 1]

                if coords[j, 2] < regions_1[n1, 2, 0]:
                    regions_1[n1, 2, 0] = coords[j, 2]

                if coords[j, 2] > regions_1[n1, 2, 1]:
                    regions_1[n1, 2, 1] = coords[j, 2]

            for k in range(3):
                regions_1[n1, k, 0] -= rep_dist2
                regions_1[n1, k, 1] += rep_dist2

            n1 += 1  # Number of primary regions

        idx_1 = regions_1[:, 0, 0].argsort(kind='heapsort').astype(np.int32)  # Sort X starts

        for a0 in range(n1):
            a = idx_1[a0]
            a1 = a * s1
            a2 = min(nCoords, a1 + s1)

            # Compare between regions
            for b0 in range(a0, n1):
                b = idx_1[b0]
                b1 = b * s1
                b2 = min(nCoords, b1 + s1)

                if regions_1[b, 0, 0] < regions_1[a, 0, 1]:

                    if regions_1[b, 1, 1] < regions_1[a, 1, 0]:
                        continue

                    if regions_1[b, 1, 0] > regions_1[a, 1, 1]:
                        continue

                    if regions_1[b, 2, 1] < regions_1[a, 2, 0]:
                        continue

                    if regions_1[b, 2, 0] > regions_1[a, 2, 1]:
                        continue

                    if a1 < b1:
                        i1 = a1
                        i2 = a2
                        j1 = b1
                        j2 = b2

                    else:
                        i1 = b1
                        i2 = b2
                        j1 = a1
                        j2 = a2

                    for i in range(i1, i2):
                        for j in range(j1, j2):
                            if j < i + 2:  # not sequential
                                continue

                            d_lim = rep_dist + radii[i] + radii[j]

                            dx = coords[i, 0] - coords[j, 0]
                            if abs(dx) > d_lim:
                                continue

                            dy = coords[i, 1] - coords[j, 1]
                            if abs(dy) > d_lim:
                                continue

                            dz = coords[i, 2] - coords[j, 2]
                            if abs(dz) > d_lim:
                                continue

                            d2 = dx * dx + dy * dy + dz * dz

                            d_lim2 = d_lim * d_lim
                            if d2 > d_lim2:
                                continue

                            if n < nRepMax:
                                rep_list[n, 0] = i
                                rep_list[n, 1] = j
                                n += 1

                            n_rep_found += 1

                else:  # All subsequent regions do not overlap
                    break

    else:  # Double split
        # Calc bounding X regions

        n1 = 0
        for i in range(0, nCoords, s2):  # Large region starts
            for a in range(3):
                regions_1[n1, a, 0] = coords[i, a]
                regions_1[n1, a, 1] = coords[i, a]

            j = i
            for n2 in range(s3):  # Small regions
                for a in range(3):
                    regions_2[n1, n2, a, 0] = coords[j, a]
                    regions_2[n1, n2, a, 1] = coords[j, a]

                # Bounding box for small region

                for k in range(j + 1, j + s1):  # All remaining points
                    if k >= nCoords:
                        break

                    for a in range(3):
                        if coords[k, a] < regions_2[n1, n2, a, 0]:
                            regions_2[n1, n2, a, 0] = coords[k, a]

                        if coords[k, a] > regions_2[n1, n2, a, 1]:
                            regions_2[n1, n2, a, 1] = coords[k, a]

                            # Bounding box of large region using small regions

                for a in range(3):
                    if regions_2[n1, n2, a, 0] < regions_1[n1, a, 0]:
                        regions_1[n1, a, 0] = regions_2[n1, n2, a, 0]

                    if regions_2[n1, n2, a, 1] > regions_1[n1, a, 1]:
                        regions_1[n1, a, 1] = regions_2[n1, n2, a, 1]

                for a in range(3):
                    regions_2[n1, n2, a, 0] -= rep_dist2
                    regions_2[n1, n2, a, 1] += rep_dist2

                j += s1

                if j >= nCoords:
                    break

            for a in range(3):
                regions_1[n1, a, 0] -= rep_dist2
                regions_1[n1, a, 1] += rep_dist2

            n1 += 1

        idx_1 = regions_1[:, 0, 0].argsort(kind='heapsort').astype(np.int32)  # Order X starts
        idx_2 = regions_2[:, :, 0, 0].argsort(kind='heapsort', axis=1).astype(np.int32)

        for p0 in range(n1):
            p = idx_1[p0]

            for q0 in range(p0, n1):
                q = idx_1[q0]

                if regions_1[q, 0, 0] >= regions_1[p, 0, 1]:  # Remainder do not overlap as they are in X order
                    break

                if regions_1[q, 1, 1] < regions_1[p, 1, 0]:
                    continue

                if regions_1[q, 1, 0] > regions_1[p, 1, 1]:
                    continue

                if regions_1[q, 2, 1] < regions_1[p, 2, 0]:
                    continue

                if regions_1[q, 2, 0] > regions_1[p, 2, 1]:
                    continue

                # Big bounding boxes overlap

                for a0 in range(s3):

                    a = idx_2[p, a0]
                    a1 = p * s2 + a * s1  # Region start
                    a2 = min(nCoords, a1 + s1)  # Region limit

                    if p == q:
                        n2 = a0  # Only compare other ones in same big box
                    else:
                        n2 = 0  # Compare all in other big box

                    # Compare between regions
                    for b0 in range(n2, s3):
                        b = idx_2[q, b0]

                        if regions_2[q, b, 0, 0] >= regions_2[p, a, 0, 1]:
                            break

                        if regions_2[q, b, 0, 1] < regions_2[p, a, 0, 0]:  # Can happen across different big boxes
                            continue

                        if regions_2[q, b, 1, 1] < regions_2[p, a, 1, 0]:
                            continue

                        if regions_2[q, b, 1, 0] > regions_2[p, a, 1, 1]:
                            continue

                        if regions_2[q, b, 2, 1] < regions_2[p, a, 2, 0]:
                            continue

                        if regions_2[q, b, 2, 0] > regions_2[p, a, 2, 1]:
                            continue

                        b1 = q * s2 + b * s1
                        b2 = min(nCoords, b1 + s1)

                        if a1 < b1:
                            i1 = a1
                            i2 = a2
                            j1 = b1
                            j2 = b2

                        else:
                            i1 = b1
                            i2 = b2
                            j1 = a1
                            j2 = a2

                        # Small bounding boxes overlap

                        for i in range(i1, i2):
                            for j in range(j1, j2):
                                if j < i + 2:  # avoid sequential
                                    continue

                                d_lim = rep_dist + radii[i] + radii[j]

                                dx = coords[i, 0] - coords[j, 0]
                                if abs(dx) > d_lim:
                                    continue

                                dy = coords[i, 1] - coords[j, 1]
                                if abs(dy) > d_lim:
                                    continue

                                dz = coords[i, 2] - coords[j, 2]
                                if abs(dz) > d_lim:
                                    continue

                                d2 = dx * dx + dy * dy + dz * dz

                                d_lim2 = d_lim * d_lim
                                if d2 > d_lim2:
                                    continue

                                if n < nRepMax:
                                    rep_list[n, 0] = i
                                    rep_list[n, 1] = j
                                    n += 1

                                n_rep_found += 1

    return n, n_rep_found


def getTemp(masses, veloc, nCoords):

    kin = 0.0
    for i in range(nCoords):
        kin += masses[i] * (veloc[i, 0] * veloc[i, 0] + veloc[i, 1] * veloc[i, 1] + veloc[i, 2] * veloc[i, 2])

    return kin / (3 * nCoords * BOLTZMANN_K)


def getStats(restIndices, restLimits, coords, nRest):

    nViol = 0
    s = 0
    for i in range(nRest):
        j = restIndices[i, 0]
        k = restIndices[i, 1]

        if j == k:
            continue
        dmin = restLimits[i, 0]
        dmax = restLimits[i, 1]

        dx = coords[j, 0] - coords[k, 0]
        dy = coords[j, 1] - coords[k, 1]
        dz = coords[j, 2] - coords[k, 2]
        r = sqrt(dx * dx + dy * dy + dz * dz)

        if r < dmin:
            viol = dmin - r
            nViol += 1
        elif r > dmax:
            viol = r - dmax
            nViol += 1
        else:
            viol = 0

        s += viol * viol

    return nViol, sqrt(s / nRest)


def updateMotion(masses, forces, accel, veloc, coords, nCoords, tRef, tStep, beta):
    rtStep = 0.5 * tStep * tStep
    temp = getTemp(masses, veloc, nCoords)
    temp = max(temp, 0.001)
    r = beta * (tRef / temp - 1.0)

    for i in range(nCoords):
        accel[i, 0] = forces[i, 0] / masses[i] + r * veloc[i, 0]
        accel[i, 1] = forces[i, 1] / masses[i] + r * veloc[i, 1]
        accel[i, 2] = forces[i, 2] / masses[i] + r * veloc[i, 2]

        coords[i, 0] += tStep * veloc[i, 0] + rtStep * accel[i, 0]
        coords[i, 1] += tStep * veloc[i, 1] + rtStep * accel[i, 1]
        coords[i, 2] += tStep * veloc[i, 2] + rtStep * accel[i, 2]

        veloc[i, 0] += tStep * accel[i, 0]
        veloc[i, 1] += tStep * accel[i, 1]
        veloc[i, 2] += tStep * accel[i, 2]

def updateVelocity(masses, forces, accel, veloc, nCoords, tRef, tStep, beta):
    temp = getTemp(masses, veloc, nCoords)
    # avoid division by 0 temperature
    temp = max(temp, 0.001)
    r = beta * (tRef / temp - 1.0)

    for i in range(nCoords):
        veloc[i, 0] += 0.5 * tStep * (forces[i, 0] / masses[i] + r * veloc[i, 0] - accel[i, 0])
        veloc[i, 1] += 0.5 * tStep * (forces[i, 1] / masses[i] + r * veloc[i, 1] - accel[i, 1])
        veloc[i, 2] += 0.5 * tStep * (forces[i, 2] / masses[i] + r * veloc[i, 2] - accel[i, 2])


def getRepulsiveForce(rep_list, forces, coords, n_rep, f_const, radii):
    total_force = 0

    if f_const == 0:
        return total_force

    for i in range(n_rep):
        j, k = rep_list[i]

        rep_dist = radii[j] + radii[k]
        rep_dist_sq = rep_dist ** 2

        dx = coords[k, 0] - coords[j, 0]
        dy = coords[k, 1] - coords[j, 1]
        dz = coords[k, 2] - coords[j, 2]

        if abs(dx) > rep_dist or abs(dy) > rep_dist or abs(dz) > rep_dist:
            continue

        d_sq = dx ** 2 + dy ** 2 + dz ** 2

        if d_sq > rep_dist_sq:
            continue

        dr = rep_dist_sq - d_sq

        # Energy contribution
        total_force += f_const * dr * dr

        rjk = 4 * f_const * dr

        dx *= rjk
        dy *= rjk
        dz *= rjk

        # Force contributions
        forces[j] -= [dx, dy, dz]
        forces[k] += [dx, dy, dz]

    return total_force


def getRestraintForce(forces, coords, restIndices, restLimits, restAmbig, nRest, fConst, exponent=2.0, distSwitch=0.5, asymptote=1.0):

    force = 0
    b = asymptote * distSwitch * distSwitch - exponent * distSwitch * distSwitch * distSwitch
    a = distSwitch * distSwitch - asymptote * distSwitch - b / distSwitch
    i = 0
    while i < nRest:

        nAmbig = restAmbig[i]
        r2 = 0.0

        if nAmbig == 1:
            j = restIndices[i, 0]
            k = restIndices[i, 1]

            if j != k:
                dx = coords[j, 0] - coords[k, 0]
                dy = coords[j, 1] - coords[k, 1]
                dz = coords[j, 2] - coords[k, 2]
                r2 = dx * dx + dy * dy + dz * dz

        else:
            for n in range(nAmbig):
                j = restIndices[i + n, 0]
                k = restIndices[i + n, 1]

                if j != k:
                    dx = coords[j, 0] - coords[k, 0]
                    dy = coords[j, 1] - coords[k, 1]
                    dz = coords[j, 2] - coords[k, 2]
                    r = dx * dx + dy * dy + dz * dz
                    r2 += 1.0 / (r * r)

            if r2 > 0:
                r2 = 1.0 / sqrt(r2)

        if r2 <= 0:
            i += nAmbig
            continue

        dmin = restLimits[i, 0]
        dmax = restLimits[i, 1]

        da = dmax + distSwitch

        if r2 < dmin * dmin:
            r2 = max(r2, 1e-8)
            r = sqrt(r2)
            d = dmin - r
            ujk = fConst * d * d
            rjk = fConst * 2 * d

        elif r2 > dmax * dmax:
            r = sqrt(r2)
            d = r - dmax

            if r <= da:
                ujk = fConst * d * d
                rjk = - fConst * 2 * d

            else:
                ujk = fConst * (a + asymptote * d + b / d)
                rjk = - fConst * (asymptote - b / (d * d))

        else:
            ujk = rjk = 0
            r = 1.0

        force += ujk

        if nAmbig == 1:
            j = restIndices[i, 0]
            k = restIndices[i, 1]

            if j == k:
                i += nAmbig
                continue

            t = rjk / r
            dx = coords[j, 0] - coords[k, 0]
            dy = coords[j, 1] - coords[k, 1]
            dz = coords[j, 2] - coords[k, 2]

            dx *= t
            dy *= t
            dz *= t

            forces[j, 0] += dx
            forces[j, 1] += dy
            forces[j, 2] += dz
            forces[k, 0] -= dx
            forces[k, 1] -= dy
            forces[k, 2] -= dz

        else:

            for n in range(nAmbig):
                j = restIndices[i + n, 0]
                k = restIndices[i + n, 1]

                if j == k:
                    continue

                dx = coords[j, 0] - coords[k, 0]
                dy = coords[j, 1] - coords[k, 1]
                dz = coords[j, 2] - coords[k, 2]

                s2 = dx * dx + dy * dy + dz * dz
                t = rjk * r2 * r2 * r / (s2 * s2 * s2)

                dx *= t
                dy *= t
                dz *= t

                forces[j, 0] += dx
                forces[k, 0] -= dx
                forces[j, 1] += dy
                forces[k, 1] -= dy
                forces[j, 2] += dz
                forces[k, 2] -= dz

        i += nAmbig

    return force

def getSupportedPairs(positions, threshold=2000000, posErr=100):
    n = len(positions)
    supported = np.zeros(n, np.int32)
    for i in range(n):
        pA = positions[i, 0]
        pB = positions[i, 1]

        for j in range(n):
            if j == i:
                continue

            pC = positions[j, 0]
            pD = positions[j, 1]

            if (posErr < abs(pC - pA) < threshold) and (posErr < abs(pD - pB) < threshold):
                supported[i] = 1
                break

            elif (posErr < abs(pD - pA) < threshold) and (posErr < abs(pC - pB) < threshold):
                supported[i] = 1
                break
    indices = supported.nonzero()[0]
    return indices


def getInterpolatedCoords(chromosomes, pos_dict, prev_pos_dict, coords):
    # Calculate index offsets for each chromosome
    offsets = {}
    total_positions = 0
    for chromo in chromosomes:
        offsets[chromo] = total_positions
        total_positions += len(pos_dict[chromo])

    prev_offsets = {}
    total_prev_positions = 0
    for chromo in chromosomes:
        prev_offsets[chromo] = total_prev_positions
        total_prev_positions += len(prev_pos_dict[chromo])

    new_coords = np.empty((total_positions, 3), float)

    for chromo, positions in pos_dict.items():
        i0 = offsets[chromo]
        j0 = prev_offsets[chromo]
        prev_positions = prev_pos_dict[chromo]
        n = len(positions)
        m = len(prev_positions)

        for i in range(n):
            # Find closest old positions for coordinate interpolation
            p1 = 0
            d_min = positions[i] - prev_positions[0]

            for j in range(1, m):
                d = positions[i] - prev_positions[j]

                if abs(d) < abs(d_min):
                    p1 = j
                    d_min = d  # closest pos

                elif abs(d) > abs(d_min):  # Seq positions were in order
                    break

            if d_min == 0:  # New position coincides with an old position
                p2 = p1
            elif d_min > 0:  # New pos is above p1
                p2 = min(p1 + 1, m - 1)
            else:  # New pos is below p1
                p2 = p1
                p1 = max(0, p1 - 1)
                d_min = positions[i] - prev_positions[p1]

            # Calculate coordinates
            if prev_positions[p2] == prev_positions[p1]:
                new_coords[i0 + i] = coords[j0 + p1]
            else:  # Interpolate
                f = float(d_min) / float(prev_positions[p2] - prev_positions[p1])
                g = 1.0 - f

                new_coords[i0 + i] = g * coords[j0 + p1] + f * coords[j0 + p2]

    return new_coords

def runDynamics(coords, masses, radii, restIndices, restLimits, restAmbig, tRef=1000.0, tStep=0.001, nSteps=1000,
                fConstR=1.0, fConstD=25.0, beta=10.0, repDist=2.0, tTaken=0.0, printInterval=10000,
                tot0=20.458, nRepMax=0):
    nRest = len(restIndices)
    nCoords = len(coords)

    if nCoords < 2:
        raise NucCythonError('Too few coodinates')

    indices = set(restIndices.ravel())
    if min(indices) < 0:
        raise NucCythonError('Restraint index negative')

    if max(indices) >= nCoords:
        data = (max(indices), nCoords)
        raise NucCythonError('Restraint index "%d" out of bounds (> %d)' % data)

    if nCoords != len(masses):
        raise NucCythonError('Masses list size does not match coordinates')

    if nRest != len(restLimits):
        raise NucCythonError('Number of restraint index pairs does not match number of restraint limits')

    if len(restAmbig) != nRest:
        raise NucCythonError('Size of ambiguity list does not match number of restraints')

    nRep = 0
    if not nRepMax:
        nRepMax = nCoords * 10
    deltaLim = 0.25 * repDist * repDist

    tStep0 = tStep * tot0
    beta /= tot0

    veloc = np.random.normal(0.0, 1.0, (nCoords, 3))
    veloc *= sqrt(tRef / getTemp(masses, veloc, nCoords))

    rep_list = np.zeros((nRepMax, 2), np.int32)
    coordsPrev = np.empty((nCoords, 3)) # typing this array important for speed
    accel = np.zeros((nCoords, 3))
    forces = np.empty((nCoords, 3))

    s1 = 8        # Small bounding box size
    s2 = 32 * s1  # Large bounding box size
    n_rep_found = None

    if nCoords >= s2 * s1:  # Double split
        s0 = s2
    else:
        s0 = s1

    regions_1 = np.zeros((1 + nCoords // s0, 3, 2))  # Large bounding boxes
    regions_2 = np.zeros((1 + nCoords // s0, s2 // s1, 3, 2))  # Small bounding boxes
    idx_1 = np.zeros(len(regions_1), np.int32)
    idx_2 = np.zeros((len(regions_1), s1), np.int32)
    max_radius = radii.max()

    for step in range(nSteps):

        if step == 0:

            nRep, n_rep_found = getRepulsionList(rep_list, coords, regions_1, regions_2, idx_1, idx_2, s1, s2, nCoords,
                                                 nRepMax, repDist, radii, max_radius)

            if n_rep_found > nRepMax:
                nRepMax = np.int32(n_rep_found * 1.1)
                rep_list = np.zeros((nRepMax, 2), np.int32)

            for i in range(nCoords):
                coordsPrev[i, 0] = coords[i, 0]
                coordsPrev[i, 1] = coords[i, 1]
                coordsPrev[i, 2] = coords[i, 2]
                forces[i, 0] = 0.0
                forces[i, 1] = 0.0
                forces[i, 2] = 0.0

            fRep = getRepulsiveForce(rep_list, forces, coords, nRep, fConstR, radii)
            fDist = getRestraintForce(forces, coords, restIndices, restLimits, restAmbig, nRest, fConstD)

            for i in range(nCoords):
                accel[i, 0] = forces[i, 0] / masses[i]
                accel[i, 1] = forces[i, 1] / masses[i]
                accel[i, 2] = forces[i, 2] / masses[i]

        else:
            maxDelta = 0.0
            for i in range(nCoords):
                dx = coords[i, 0] - coordsPrev[i, 0]
                dy = coords[i, 1] - coordsPrev[i, 1]
                dz = coords[i, 2] - coordsPrev[i, 2]
                d2 = dx * dx + dy * dy + dz * dz

                if d2 > maxDelta:
                    maxDelta = d2

                    if maxDelta > deltaLim:
                        break

            if maxDelta > deltaLim:
                nRep, n_rep_found = getRepulsionList(rep_list, coords, regions_1, regions_2, idx_1, idx_2, s1, s2,
                                                     nCoords, nRepMax, repDist, radii, max_radius)  # Handle errors

                if n_rep_found > nRepMax:
                    nRepMax = np.int32(n_rep_found * 1.1)
                    rep_list = np.zeros((nRepMax, 2), np.int32)

                for i in range(nCoords):
                    coordsPrev[i, 0] = coords[i, 0]
                    coordsPrev[i, 1] = coords[i, 1]
                    coordsPrev[i, 2] = coords[i, 2]

        updateMotion(masses, forces, accel, veloc, coords, nCoords, tRef, tStep0, beta)

        for i in range(nCoords):
            forces[i, 0] = 0.0
            forces[i, 1] = 0.0
            forces[i, 2] = 0.0

        fRep = getRepulsiveForce(rep_list, forces, coords, nRep, fConstR, radii)
        fDist = getRestraintForce(forces, coords, restIndices, restLimits, restAmbig, nRest, fConstD)

        updateVelocity(masses, forces, accel, veloc, nCoords, tRef, tStep0, beta)

        if (printInterval > 0) and step % printInterval == 0:
            temp = getTemp(masses, veloc, nCoords)
            nViol, rmsd = getStats(restIndices, restLimits, coords, nRest)

            data = (temp, fRep, fDist, rmsd, nViol, nRep)
            print('temp:%7.2lf  fRep:%7.2lf  fDist:%7.2lf  rmsd:%7.2lf  nViol:%5d  nRep:%5d' % data)

        tTaken += tStep

    return tTaken, n_rep_found


def calc_restraints(chromosomes, contact_dict, particle_size=10000,
                    scale=1.0, exponent=-0.33,
                    lower=0.8, upper=1.2,
                    min_count=1):
    """
    Function to convert single-cell contact data into distance restraints
    for structure calculations.
    """
    num_contacts_dict = {} # Total num contacts for each chromosomes
    pos_dict = {}          # Start sequence position for each particle in each chromosome
    restraint_dict = {}    # Final restraints for each pair of chromosomes

    chromos = set(chromosomes)
    nc = len(chromosomes)

    # Num contacts per chromo for mem allocations
    for chrA in contact_dict:
        if chrA not in chromos:
            continue

        for chrB in contact_dict[chrA]:
            if chrB not in chromos:
                continue

            contacts = contact_dict[chrA][chrB]
            n = len(contacts[0])

            if chrA in num_contacts_dict:
                num_contacts_dict[chrA] += n
            else:
                num_contacts_dict[chrA] = n

            if chrB in num_contacts_dict:
                num_contacts_dict[chrB] += n
            else:
                num_contacts_dict[chrB] = n

    chromo_idx = {chromo:i for i, chromo in enumerate(chromosomes)}
    limits = np.zeros((nc,2), dtype=np.int32)

    # Get chromosome ranges
    for chrA in contact_dict:
        if chrA not in chromos:
            continue

        a = chromo_idx[chrA]

        for chrB in contact_dict[chrA]:
            if chrB not in chromos:
                continue

            b = chromo_idx[chrB]
            contacts = contact_dict[chrA][chrB]
            n = len(contacts[0])

            for i in range(n):

                if limits[a,0] == 0: # zero is not a valid seq pos anyhow
                    limits[a,0] = contacts[0,i]
                elif contacts[0,i] < limits[a,0]:
                    limits[a,0] = contacts[0,i]

                if limits[b,0] == 0:
                    limits[b,0] = contacts[1,i]
                elif contacts[1,i] < limits[b,0]:
                    limits[b,0] = contacts[1,i]

                if contacts[0,i] > limits[a,1]:
                    limits[a,1] = contacts[0,i]

                if contacts[1,i] > limits[b,1]:
                    limits[b,1] = contacts[1,i]


    # Shift extremities of chromosome ranges to lie exactly on particle region edges
    for a in range(nc):
        limits[a,0] = particle_size * (limits[a,0]//particle_size)
        limits[a,1] = particle_size * int(math.ceil(limits[a,1]/particle_size))

    # Get particle region start positions for each chromosome
    for chromo in chromosomes:
        a = chromo_idx[chromo]
        pos_dict[chromo] = np.arange(limits[a,0], limits[a,1] + particle_size, particle_size, dtype=np.int32)

    # Get restraint indices, do binning of contact observations
    for chrA in contact_dict:
        if chrA not in chromos:
            continue

        a = chromo_idx[chrA]
        na = len(pos_dict[chrA])
        seq_pos_a = pos_dict[chrA]
        restraint_dict[chrA] = {}

        for chrB in contact_dict[chrA]:
            if chrB not in chromos:
                continue

            b = chromo_idx[chrB]
            nb = len(pos_dict[chrB])
            seq_pos_b = pos_dict[chrB]

            contacts = contact_dict[chrA][chrB]
            n = len(contacts[0])

            restraints = np.empty((6, n), float)
            bin_matrix = np.zeros((na, nb), dtype=np.int32)

            for i in range(n):

                # Find bin index for chromo A
                for j in range(na):
                    if seq_pos_a[j] >= contacts[0,i]:
                        break

                else:
                    continue

                # Find bin index for chromo B
                for k in range(nb):
                    if seq_pos_b[k] >= contacts[1,i]:
                        break

                else:
                    continue

                bin_matrix[j,k] += contacts[2,i]

            # loop over all binned contacts, and calculate the constraint target distance
            # using a powerlaw function and the number of observations
            k = 0
            for i in range(na):
                for j in range(nb):
                    if bin_matrix[i,j] > 0:
                        if bin_matrix[i,j] < min_count:
                            continue

                        dist = scale * bin_matrix[i,j] ** exponent

                        restraints[0,k] = float(i) #binA
                        restraints[1,k] = float(j) #binB
                        restraints[2,k] = 1.0 # Weighting not currently used
                        restraints[3,k] = dist # target value
                        restraints[4,k] = dist * lower # constraint lower bound
                        restraints[5,k] = dist * upper # constraint upper bound

                        k += 1

            restraint_dict[chrA][chrB] = restraints[:,:k]

    return restraint_dict, pos_dict


def concatenate_restraints(restraint_dict, pos_dict, particle_size,
                           backbone_lower=0.1, backbone_upper=1.1):
    """
    Joins restraints stored in a dict by chromo pairs into long concatenated arrays.
    Indices of restraints relate to concatenated chromo seq pos.
    Add-in all the backbone restraints for sequential particles.
    """

    num_restraints = 0
    chromo_idx_offset = {}

    for chr_a in sorted(pos_dict):
        n = len(pos_dict[chr_a])
        chromo_idx_offset[chr_a] = num_restraints
        num_restraints += n - 1  # One fewer because no link at end of chain

    for chr_a in restraint_dict:
        for chr_b in restraint_dict[chr_a]:
            num_restraints += restraint_dict[chr_a][chr_b].shape[1]

    # Final arrays which will hold identities of restrained particle pairs
    # and the restraint distances for each
    particle_indices = np.empty((num_restraints, 2), dtype=np.int32)
    distances = np.empty((num_restraints, 2), dtype=float)

    # Add backbone path restraints
    m = 0
    for chr_a in pos_dict:
        positions = pos_dict[chr_a]
        n = len(positions)
        start_a = chromo_idx_offset[chr_a]

        for i in range(n - 1):
            # Normally this is 1.0 for regular sized particles
            dist = (positions[i + 1] - positions[i]) / particle_size

            particle_indices[m, 0] = i + start_a
            particle_indices[m, 1] = i + start_a + 1

            distances[m, 0] = dist * backbone_lower  # lower
            distances[m, 1] = dist * backbone_upper  # upper

            m += 1

    # Add regular restraints for chromo pairs
    for chr_a in restraint_dict:
        start_a = chromo_idx_offset[chr_a]  # Offset for chromo A particles

        for chr_b in restraint_dict[chr_a]:
            start_b = chromo_idx_offset[chr_b]  # Offset for chromo B particles

            restraints = restraint_dict[chr_a][chr_b]
            n = restraints.shape[1]

            for i in range(n):
                particle_indices[m, 0] = int(restraints[0, i]) + start_a
                particle_indices[m, 1] = int(restraints[1, i]) + start_b

                distances[m, 0] = restraints[4, i]  # lower
                distances[m, 1] = restraints[5, i]  # upper

                m += 1

    return particle_indices, distances











import click
from sc3dg.commands.count import count
from sc3dg.commands.model import model
from sc3dg.commands.impute import impute


@click.group()
def cli():
    pass


cli.add_command(count)
cli.add_command(model)
cli.add_command(impute)

if __name__ == '__main__':
    cli()



import pandas as pd
import os
import cooler
import numpy as np
import tqdm
from scipy.stats import norm
import cooltools
import warnings

warnings.filterwarnings("ignore")
from sc3dg.analysis.embedding import scale_matrix, gaussian_distance, cosine_similarity, PCAEmbedding, MDSEmbedding, tSNEEmbedding, \
    UMAPEmbedding, LLEEmbedding, IsomapEmbedding
import bioframe
import sys
import sc3dg.analysis.genome as genome
from scipy.special import comb, perm
import random
from cooltools import insulation
import random
import tqdm
import shutil
import time
from joblib import Parallel, delayed
hg38_cens = pd.DataFrame({'chrom': ['chr1', 'chr10', 'chr11', 'chr12', 'chr13', 'chr14', 'chr15', 'chr16', 'chr17',
                                    'chr18', 'chr19', 'chr2', 'chr20', 'chr21', 'chr22', 'chr3', 'chr4', 'chr5', 'chr6',
                                    'chr7', 'chr8', 'chr9', 'chrX', 'chrY'],
                          'start': [122026459, 39686682, 51078348, 34769407, 16000000, 16000000, 17083673, 36311158,
                                    22813679, 15460899, 24498980, 92188145, 26436232, 10864560, 12954788, 90772458,
                                    49712061, 46485900, 58553888, 58169653, 44033744, 43389635, 58605579, 10316944],
                          'end': [124932724, 41593521, 54425074, 37185252, 18051248, 18173523, 19725254, 38265669,
                                  26616164, 20861206, 27190874, 94090557, 30038348, 12915808, 15054318, 93655574,
                                  51743951, 50059807, 59829934, 61528020, 45877265, 45518558, 62412542, 10544039],
                          'mid': [123479591, 40640101, 52751711, 35977329, 17025624, 17086761, 18404463, 37288413,
                                  24714921, 18161052, 25844927, 93139351, 28237290, 11890184, 14004553, 92214016,
                                  50728006, 48272853, 59191911, 59848836, 44955504, 44454096, 60509060, 10430491]})

mm10_cens = pd.DataFrame(columns=['chrom', 'start', 'end', 'mid'])
mm10_cens['start'] = mm10_cens['start'].astype('int64')
mm10_cens['end'] = mm10_cens['end'].astype('int64')
mm10_cens['mid'] = mm10_cens['mid'].astype('int64')

mm10_chromsize = pd.DataFrame({'chrom': ['chr1', 'chr2', 'chr3', 'chr4', 'chr5', 'chr6', 'chr7', 'chr8', 'chr9',
                                         'chr10', 'chr11', 'chr12', 'chr13', 'chr14', 'chr15', 'chr16', 'chr17',
                                         'chr18', 'chr19', 'chrX', 'chrY', 'chrM'],
                               'length': [195471971, 182113224, 160039680, 156508116, 151834684, 149736546, 145441459,
                                          129401213, 124595110, 130694993, 122082543, 120129022, 120421639, 124902244,
                                          104043685, 98207768, 94987271, 90702639, 61431566, 171031299, 91744698,
                                          16299]})
hg38_chromsize = pd.DataFrame({'chrom': ['chr1', 'chr2', 'chr3', 'chr4', 'chr5', 'chr6', 'chr7', 'chr8', 'chr9',
                                         'chr10', 'chr11', 'chr12', 'chr13', 'chr14', 'chr15', 'chr16', 'chr17',
                                         'chr18', 'chr19', 'chr20', 'chr21', 'chr22', 'chrX', 'chrY', 'chrM'],
                               'length': [248956422, 242193529, 198295559, 190214555, 181538259, 170805979, 159345973,
                                          145138636, 138394717, 133797422, 135086622, 133275309, 114364328, 107043718,
                                          101991189, 90338345, 83257441, 80373285, 58617616, 64444167, 46709983,
                                          50818468, 156040895, 57227415, 16569]})
data_info = {'hg38': [hg38_chromsize, hg38_cens], 'mm10': [mm10_chromsize, mm10_cens]}


class MMCooler:
    def __init__(self,
                 directory,  # 目录
                 resolution,  # 分辨率
                 describe=None,  # 技术名称
                 merge=True,  # 是否将读取的cell全部merge
                 genome: str = 'mm10',
                 balance: bool = True,
                 read_cache=True,
                 n_jobs=6,
                 ):
        '''
            directory:
            resolution:如果int,那没什么，就是直接读取；如果是list，list第一个必须要有，切是最小的，后面的分辨率会按照这个去合并或者读取
            describe:
            merge:
            genome:
            balance
        '''
        self.directory = directory
        self.n_jobs = n_jobs
        self.mcool_files = []
        self.mcool_names = []
        self.cache_file = []
        self.clr_obj = None

        self.__merge = merge
        self.genome = genome
        self.__balance = balance

        self.resolution = resolution if isinstance(resolution, list) else [resolution]
        self.describe = describe
        self.chromosome = ['chr1', 'chr2', 'chr3', 'chr4', 'chr5', 'chr6',
                           'chr7', 'chr8', 'chr9', 'chr10', 'chr11', 'chr12',
                           'chr13', 'chr14', 'chr15', 'chr16', 'chr17', 'chr18',
                           'chr19', 'chr20', 'chr21', 'chr22', 'chrX', 'chrY', 'chrM']
        self.__chrom_terr = ['chr1', 'chr2', 'chr3', 'chr4', 'chr5', 'chr6',
                             'chr7', 'chr8', 'chr9', 'chr10', 'chr11', 'chr12',
                             'chr13', 'chr14', 'chr15', 'chr16', 'chr17', 'chr18',
                             'chr19']
        self.read_cache = read_cache
        self.__pool = False

        if not os.path.exists('./cache'):
            os.makedirs('./cache')
        self.extract_mcool_files()
        self.read_mcool_files()
        if not os.path.exists('./file'):
            os.makedirs('./file')

    def stat_merge(self):
        return self.__merge

    def stat_balance(self):
        return self.__balance

    def stat_resolution(self):
        return isinstance(self.resolution, list)

    def shape(self):
        shape = np.array(self.clr_obj)
        return shape

    def stat_file_name(self):
        return self.mcool_names

    def __read_zoomify(self, mcool_file, mcool_name,cache_name):
        raw_res = self.resolution[0] 
        if not os.path.exists(cache_name):
            cooler.zoomify_cooler(base_uris=mcool_file + "::resolutions/" + str(raw_res), outfile=cache_name,
                                    resolutions=self.resolution,
                                    chunksize=10000, nproc=16)

    # 读取目录下所有的mcool文件
    def extract_mcool_files(self):
        if self.directory.endswith('.mcool'):
            self.mcool_files.append(self.directory)
            self.mcool_names.append(self.directory.split('/')[-1].split('.')[0])

        for root, dir_, files in os.walk(self.directory):
            for f in files:
                if f.endswith('.mcool'):
                    self.mcool_files.append(root + '/' + f)
                    self.mcool_names.append(f.split('.')[0])

    # 加载所有mcool文件到cooler.Cooler对象,如果merge不是none，那就merge到这个
    def read_mcool_files(self):
        '''
            self.coolers: list of cooler objects
        '''
        # 如果合并
        if self.__merge:
            print('run merge mode')
            # 多个分辨率
            self.clr_obj = []
            raw_res = self.resolution[0]  # 读取有这个分辨率,并以此分辨率座位基础

            # 先把每个样本bin化，然后再读取
            print('binning the raw mcool files')
            for mcool_file, mcool_name in zip(self.mcool_files, self.mcool_names):
                # 如果美有cache，那就创建，读取
                # 创建对应样本cache的名字
                cache_name = './cache/%s_multi_res_%s' % (self.describe, mcool_name)
                for res in self.resolution:
                    cache_name += '-'
                    cache_name += str(res)
                cache_name += '.mcool'

                # 如果没有这个文件，那就生成
                if not os.path.exists(cache_name):
                    cooler.zoomify_cooler(base_uris=mcool_file + "::resolutions/" + str(raw_res), outfile=cache_name,
                                          resolutions=self.resolution,
                                          chunksize=10000, nproc=32)

            # 都存到cache了，然后按照分辨率读取,并 merge
            print('reading the binned mcool files')
            for res in self.resolution:
                tmp = []
                for mcool_file, mcool_name in zip(self.mcool_files, self.mcool_names):
                    cache_name = './cache/%s_multi_res_%s' % (self.describe, mcool_name)
                    for rr in self.resolution:
                        cache_name += '-'
                        cache_name += str(rr)
                    cache_name += '.mcool'
                    tmp.append(cache_name + '::resolutions/' + str(res))

                print('merging the binned mcool files/resolution=%s' % res)
                if not os.path.exists('./cache/%s_merge_all_res-%s' % (self.describe, res)):
                    cooler.merge_coolers(input_uris=tmp,
                                         output_uri='./cache/%s_merge_all_res-%s' % (self.describe, res),
                                         mergebuf=1000000)
                self.clr_obj.append(cooler.Cooler('./cache/%s_merge_all_res-%s' % (self.describe, res)))

        # 如果不合并
        else:
            print('run nomerge mode')
            self.clr_obj = []
            # 多个分辨率

            # 先把每个样本bin化，然后再读取
            raw_res = self.resolution[0]  # 读取有这个分辨率,并以此分辨率座位基础
            print('binning the raw mcool files')
            task = []
            for mcool_file, mcool_name in zip(self.mcool_files, self.mcool_names):
                cache_name = './cache/%s_multi_nomerge_%s_res' % (self.describe, mcool_name)
                for res in self.resolution:
                    cache_name += '-'
                    cache_name += str(res)
                cache_name += '.mcool'
                task.append([ mcool_file, mcool_name, cache_name])
            Parallel(n_jobs=self.n_jobs)(delayed(self.__read_zoomify)(*t) for t in task)
            
            # 都存到cache了，然后按照分辨率读取
            # 外层是分辨率，里层是样本
            print('reading the binned mcool files')
            for res in self.resolution:
                print(res)
                tmp = []
                for mcool_file, mcool_name in zip(self.mcool_files, self.mcool_names):
                    cache_name = './cache/%s_multi_nomerge_%s_res' % (self.describe, mcool_name)
                    for r in self.resolution:
                        cache_name += '-'
                        cache_name += str(r)
                    cache_name += '.mcool'
                    tmp.append(cooler.Cooler(cache_name + '::resolutions/' + str(res)))
                    if cache_name not in self.cache_file:
                        self.cache_file.append(cache_name)
                self.clr_obj.append(tmp)

    # 获取单个细胞所有contact数量
    def get_total_contact(self, res=None):
        '''
            res: resolution
        '''
        if res is None:
            clr = self.clr_obj
        else:
            clr = [self.clr_obj[self.resolution.index(res)]]

        res = []
        for cr in clr[0]:
            res.append(cr.info['sum'])
        df = pd.DataFrame(res, index=self.mcool_names, columns=['total_contact'])
        df.to_csv('./file/{}_total_contact.csv'.format(self.describe), sep='\t')
        return res

    # 输入一个cooler，返回一组结果
    def __get_contact_vs_distance(self, clr, res, arms):
        resolution = res
        arms = arms[arms.chrom.isin(clr.chromnames)].reset_index(drop=True)
        if not cooltools.lib.is_cooler_balanced(clr):
            cooler.balance_cooler(clr, store=True)
        cvd_smooth_agg = cooltools.expected_cis(
            clr=clr,
            view_df=arms,
            smooth=True,
            aggregate_smoothed=True,
            nproc=16
        )

        cvd_smooth_agg['s_bp'] = cvd_smooth_agg['dist'] * resolution
        cvd_smooth_agg['balanced.avg.smoothed.agg'].loc[cvd_smooth_agg['dist'] < 2] = np.nan
        cvd_smooth_agg = cvd_smooth_agg.groupby('s_bp').sum().reset_index()
        cvd_smooth_agg = cvd_smooth_agg.set_index('s_bp')

        return cvd_smooth_agg.loc[:, ['balanced.avg.smoothed.agg']]

    # 获得绘制P（s）图像的数据
    def get_contact_vs_distance(self, res=None, version='hg38'):
        '''
            dismiss:舍弃对角线开始的第几个细胞
            mode:
                cis:   interaction decay with distance
                 or trans:teraction frequencies for inter-chromosomal blocks
            intra_only:  Return expected only for symmetric intra-regions defined by view_df,
            i.e. chromosomes, chromosomal-arms, intra-domains, etc.
            When False returns expected both for symmetric intra-regions and
            assymetric inter-regions.
        '''

        if res is None:
            sys.exit('res is None')

        chromsizes, cens = data_info[version]
        arms = bioframe.make_chromarms(chromsizes, cens)

        clr = self.clr_obj[self.resolution.index(res)]
        result = []
        for name, val in zip(self.mcool_names, clr):
            tmp = self.__get_contact_vs_distance(val, res, arms)
            tmp.columns = [name]
            result.append(tmp)
        result = pd.concat(result, axis=1)

        return result

    # 获得intra-chrom contact个数
    def get_intra_contact_count(self, split_chorm: bool = True):
        '''
            merge + split: 返回染色体dict
            merge + nosolt:返回一个值
            nomerge + nosplit:返回list of count
            nomerge + split:返回 list of dict
        '''

        counts = []

        for val in self.clr_obj[-1]:
            counts.append(self.__get_intra_contact_count(val, split_chorm))

        df = pd.DataFrame(counts)
        df.index = self.mcool_names

        df.to_csv('./file/{}_splitChrom:{}_intra_contact_count.csv'.format(self.describe, split_chorm), sep='\t')

        return counts

    # 一个cool，一个dict
    def __get_intra_contact_count(self, clr, split_chorm: bool = True):

        bins = clr.bins()[:]
        pixels = clr.pixels()[:]
        pixels['chr1'] = list(bins.iloc[list(pixels['bin1_id']), 0])
        pixels['chr2'] = list(bins.iloc[list(pixels['bin2_id']), 0])
        pixels = pixels[pixels['chr1'] == pixels['chr2']]
        if split_chorm:
            # 如果merge，那就返回一个dict
            counts = {}
            for chrom in clr.chromnames:
                counts[chrom] = pixels[pixels['chr1'] == chrom].shape[0]
        # 不分染色体，那就只返回一个数值
        else:
            counts = {'total': 0}
            for chrom in clr.chromnames:
                counts['total'] += pixels[pixels['chr1'] == chrom].shape[0]

        return counts

    # 获得contact agg,只算上三角+对角线，求和
    def get_intra_contact_agg(self, split_chorm: bool = True, res=None):
        '''
            merge + split: 返回染色体dict
            merge + nosolt:返回一个值
            nomerge + nosplit:返回list of count
            nomerge + split:返回 list of dict
        '''
        counts = []

        for val in self.clr_obj[-1]:
            counts.append(self.__get_intra_contact_agg(val, split_chorm))

        df = pd.DataFrame(counts)
        df.index = self.mcool_names

        df.to_csv('./file/{}_splitChrom:{}_intra_contact_agg.csv'.format(self.describe, split_chorm), sep='\t')
        return counts

    # 一个cooelr，一个result
    def __get_intra_contact_agg(self, clr, split_chorm: bool = True):
        # 分染色体计数
        bins = clr.bins()[:]
        pixels = clr.pixels()[:]
        pixels['chr1'] = list(bins.iloc[list(pixels['bin1_id']), 0])
        pixels['chr2'] = list(bins.iloc[list(pixels['bin2_id']), 0])
        pixels['start'] = list(bins.iloc[list(pixels['bin1_id']), 1])
        pixels['end'] = list(bins.iloc[list(pixels['bin1_id']), 2])
        pixels = pixels[pixels['chr1'] == pixels['chr2']]
        if split_chorm:
            # 如果merge，那就返回一个dict

            counts = {}
            for chrom in self.chromosome:
                try:
                    tmp = pixels[(pixels['chr1'] == chrom)]
                except:
                    continue

                counts[chrom] = sum(tmp['count'])
        # 不分染色体，那就只返回一个数值
        else:

            counts = {'total': sum(pixels[pixels['chr1'] == pixels['chr2']]['count'])}

        return counts

    # 获得inter-chrom contact个数
    def get_inter_contact_count(self):
        counts = []

        for val in self.clr_obj[-1]:
            counts.append(self.__get_inter_contact_count(val))
        df = pd.DataFrame(counts)
        df.index = self.mcool_names

        df.to_csv('./file/{}_inter_contact_count.csv'.format(self.describe), sep='\t')
        return counts

    def __get_inter_contact_count(self, clr):
        # 分染色体计数
        bins = clr.bins()[:]
        pixels = clr.pixels()[:]
        pixels['chr1'] = list(bins.iloc[list(pixels['bin1_id']), 0])
        pixels['chr2'] = list(bins.iloc[list(pixels['bin2_id']), 0])
        pixels['start'] = list(bins.iloc[list(pixels['bin1_id']), 1])
        pixels['end'] = list(bins.iloc[list(pixels['bin1_id']), 2])
        pixels = pixels[pixels['chr1'] != pixels['chr2']]
        counts = {'total': pixels.shape[0]}

        return counts

    # 获得inter-chrom contact总数
    def get_inter_contact_agg(self, split_chrom: bool = True, res=None):
        counts = []

        for val in self.clr_obj[-1]:
            counts.append(self.__get_inter_contact_agg(val))
        df = pd.DataFrame(counts)
        df.index = self.mcool_names

        df.to_csv('./file/{}_inter_contact_agg.csv'.format(self.describe), sep='\t')
        return counts

    def __get_inter_contact_agg(self, clr, split_chorm: bool = True):
        # 分染色体计数
        bins = clr.bins()[:]
        pixels = clr.pixels()[:]
        pixels['chr1'] = list(bins.iloc[list(pixels['bin1_id']), 0])
        pixels['chr2'] = list(bins.iloc[list(pixels['bin2_id']), 0])
        pixels['start'] = list(bins.iloc[list(pixels['bin1_id']), 1])
        pixels['end'] = list(bins.iloc[list(pixels['bin1_id']), 2])
        pixels = pixels[pixels['chr1'] != pixels['chr2']]
        counts = {'total': sum(pixels['count'])}

        return counts

    # 获得clr_obj对象中的embeeding,仅在merge=False时候使用
    def get_clustering_embedding(self, res=None, cache=True,
                                 similarityMethod='gaussian_distance',
                                 embeddingMethod='UMAP',
                                 dim=2):
        if self.__merge:
            print('merged,no embedding')
            return

        if res is None:
            print('please input resolution')
            return

        CellSimilarity = []
        CellFeature = []

        clr = self.clr_obj[self.resolution.index(res)]

        fileCount = len(clr)

        if not cache or not os.path.exists('./cache/cache_%s_%s.npz' % (self.describe, res)):
            # 不存在cache，那还是要做一遍，但是存下来

            for cr in clr:
                ####获取该细胞的HIC矩阵
                TargetMatrix_1 = cr.matrix(balance=False)[:]

                ###获取距离对角线0-10的数据并降成一维
                diagonal_data = [np.diag(TargetMatrix_1, k) for k in range(0, 10)]

                ###获取降维后的矩阵
                SampledTargetMatrix_1 = scale_matrix(TargetMatrix_1, 200)

                ###整合整体特征与染色体内部特征
                CellFeature.append(np.concatenate(
                    [np.concatenate([arr.flatten() for arr in diagonal_data]), SampledTargetMatrix_1.flatten()],
                    axis=0))

                ####计算细胞间的similarity
                CurCellSimilarity = []
                for j in range(fileCount):
                    TargetMatrix_2 = clr[j].matrix(balance=False)[:]
                    SampledTargetMatrix_2 = scale_matrix(TargetMatrix_2, 200)
                    if similarityMethod == 'gaussian_distance':
                        CurCellSimilarity.append(gaussian_distance(SampledTargetMatrix_1, SampledTargetMatrix_2, 1))
                    if similarityMethod == 'EuclideanDistance':
                        CurCellSimilarity.append(np.linalg.norm(SampledTargetMatrix_1 - SampledTargetMatrix_2))
                    if similarityMethod == 'CosineSimilarity':
                        CurCellSimilarity.append(1 - cosine_similarity(SampledTargetMatrix_1, SampledTargetMatrix_2))
                ###记录每个细胞与所有细胞间的similarity
                CellSimilarity.append(CurCellSimilarity)
        else:
            ###存在cache，那就直接读取
            print('load cache')
            cache = np.load('./cache/cache_%s_%s.npz' % (self.describe, res))
            CellFeature = cache['CellFeature']
            CellSimilarity = cache['CellSimilarity']

        # print(CellFeature)
        ##对矩阵正态分布处理
        CellSimilarity = np.array(CellSimilarity)
        smoothed_CellSimilarity = 1 - norm.cdf(CellSimilarity, np.mean(CellSimilarity), np.std(CellSimilarity))
        smoothed_CellSimilarity = (smoothed_CellSimilarity - np.min(smoothed_CellSimilarity)) / (
                    np.max(smoothed_CellSimilarity) - np.min(smoothed_CellSimilarity))

        ###计算每个细胞的embadding
        CellFeature = np.array(CellFeature)
        self.CellFeature = CellFeature
        self.cellSimilarity = smoothed_CellSimilarity

        if cache and not os.path.exists('./cache/cache_%s_%s.npz' % (self.describe, res)):
            if not os.path.exists('cache'):
                os.makedirs('cache')
            print('save cache')
            cache = {}
            cache['CellFeature'] = CellFeature
            cache['CellSimilarity'] = smoothed_CellSimilarity
            cache['mcool_names'] = self.mcool_names

            np.savez('./cache/cache_%s_%s.npz' % (self.describe, res), **cache)

        print(CellFeature)
        if embeddingMethod == 'PCA':
            CellsEmbedding = PCAEmbedding(CellFeature, dim=dim)

        if embeddingMethod == 'MDS':
            CellsEmbedding = MDSEmbedding(CellFeature, dim=dim)

        if embeddingMethod == 'tSNE':
            CellsEmbedding = tSNEEmbedding(CellFeature, dim=dim)

        if embeddingMethod == 'UMAP':
            CellsEmbedding = UMAPEmbedding(CellFeature, dim=dim)

        if embeddingMethod == 'LLE':
            CellsEmbedding = LLEEmbedding(CellFeature, dim=dim)

        if embeddingMethod == 'Isomap':
            CellsEmbedding = IsomapEmbedding(CellFeature, dim=dim)
        else:
            print('please input right embedding method')
            return smoothed_CellSimilarity

        return smoothed_CellSimilarity, CellsEmbedding

    # 获得每个cell的对角线元素flateen的feature
    def get_clustering_diagnal(self, dim=10, res=None):

        shape = np.array(self.clr_obj).shape

        if len(shape) == 1:
            CellFeature = []
            for clr in self.clr_obj:
                ####获取该细胞的HIC矩阵
                TargetMatrix_1 = clr.matrix(balance=False)[:]

                ###获取距离对角线0-10的数据并降成一维
                diagonal_data = [np.diag(TargetMatrix_1, k) for k in range(0, dim)]

                ###获取降维后的矩阵
                SampledTargetMatrix_1 = scale_matrix(TargetMatrix_1, 200)

                ###整合整体特征与染色体内部特征
                result = np.concatenate(
                    [np.concatenate([arr.flatten() for arr in diagonal_data]), SampledTargetMatrix_1.flatten()], axis=0)
                CellFeature.append(result)
        else:
            CellFeature = []
            for lt in self.clr_obj:
                tmp_feature = []
                for clr in lt:
                    ####获取该细胞的HIC矩阵
                    TargetMatrix_1 = clr.matrix(balance=True)[:]

                    ###获取距离对角线0-10的数据并降成一维
                    diagonal_data = [np.diag(TargetMatrix_1, k) for k in range(0, dim)]

                    ###获取降维后的矩阵
                    SampledTargetMatrix_1 = scale_matrix(TargetMatrix_1, 200)

                    ###整合整体特征与染色体内部特征
                    result = np.concatenate(
                        [np.concatenate([arr.flatten() for arr in diagonal_data]), SampledTargetMatrix_1.flatten()],
                        axis=0)
                    tmp_feature.append(result)
                CellFeature.append(tmp_feature)
        return CellFeature

    # 获得每个mcool文件的contact per bin
    def get_avg_contact_per_bin(self, ignore_diag=False, res=None):
        if res is None:
            clr = self.clr_obj
        else:
            clr = self.clr_obj[self.resolution.index(res)]

        shape = np.array(clr).shape

        if len(shape) == 1:
            avgs = []
            for cr in clr:
                avgs.append(self.__get_avg_contact_per_bin(cr, ignore_diag))
        else:
            avgs = []
            for val in clr:
                tmp_avg = []
                for cr in val:
                    tmp_avg.append(self.__get_avg_contact_per_bin(cr, ignore_diag))
                avgs.append(tmp_avg)

        return avgs

    def __get_avg_contact_per_bin(self, clr, ignore_diag=False):
        top = 0
        down = 0
        for chrom in self.chromosome:

            try:
                mat = clr.matrix(balance=False).fetch(chrom)
            except:
                continue
            if ignore_diag:
                top += np.triu(mat, 1).sum()
            else:
                top += np.triu(mat).sum()

            down += mat.shape[0]

        return top / down

    # 获得不同分辨的覆盖率
    def get_coverage(self, res=None):

        if res is None:
            clr = self.clr_obj
        else:
            clr = [self.clr_obj[self.resolution.index(res)]]

        coverages = []
        for val in clr:
            tmp_coverage = []
            for cr in val:
                tmp_coverage.append(self.__get_coverage(cr))
            coverages.append(tmp_coverage)

        if res is None:
            for res in self.resolution:
                df = pd.DataFrame(coverages[self.resolution.index(res)])
                df.index = self.mcool_names
                df.to_csv('./file/{}_{}_coverages.csv'.format(self.describe, res), sep='\t')

        else:
            df = pd.DataFrame(coverages[0])
            df.index = self.mcool_names
            df.to_csv('./file/{}_{}_coverages.csv'.format(self.describe, res), sep='\t')

        return coverages

    def __get_coverage(self, clr):

        bins = clr.bins()[:]
        pixels = clr.pixels()[:]
        cover = set(set(np.unique(pixels['bin1_id'])).union(set(np.unique(pixels['bin2_id']))))
        freq = len(cover) / bins.shape[0]
        return freq

    # 获得累计覆盖率
    def get_acum_coverage(self, res=None):
        if self.__merge:
            sys.exit('merge is True, can not get acum_coverage')

        if res is None:
            clr = self.clr_obj
        else:
            clr = [self.clr_obj[self.resolution.index(res)]]

        coverages = []
        for val in clr:
            tmp_coverage = []
            for cr in val:
                tmp_coverage.append(self.__get_acum_coverage(cr))
            coverages.append(tmp_coverage)

        return coverages

    def __get_acum_coverage(self, clr):
        freq = []
        bins = clr[0].bins()[:]
        pixels = clr[0].pixels()[:]
        cover = set(set(np.unique(pixels['bin1_id'])).union(set(np.unique(pixels['bin2_id']))))
        freq.append(len(cover) / bins.shape[0])
        for cr in clr[1:]:
            bins = cr.bins()[:]
            pixels = cr.pixels()[:]
            cover = cover.union(set(np.unique(pixels['bin1_id'])).union(set(np.unique(pixels['bin2_id']))))
            freq.append(len(cover) / bins.shape[0])
        freq = sorted(freq)
        return freq

    # 计算染色质疆域的nDS
    def get_nDS_chrom_terr(self, res=None, epoch=100):
        if self.__merge:
            sys.exit('merge is True, can not get nDS_chrom_terr_per_cell')

        if res is None:
            clr = self.clr_obj
        else:
            clr = [self.clr_obj[self.resolution.index(res)]]

        self.chrom_pair = []
        for cr in self.__chrom_terr:
            for cr2 in self.__chrom_terr[self.__chrom_terr.index(cr) + 1:]:
                self.chrom_pair.append([cr, cr2])

        result = []
        for i, val in enumerate(clr):
            # print(i)
            result.append(self.__get_nDS_chrom_terr(val, i, epoch))

        return result

    # 获得单个细胞的每个pair的nDS和result
    def __get_nDS_chrom_terr(self, clr, idx, epoch):
        print('cal DS')
        DS = []
        for cell in clr:
            bins = cell.bins()[:]
            pair_DS = []
            pixels = cell.pixels()[:]
            # print(bins)
            pixels['chr1'] = list(bins.iloc[list(pixels['bin1_id']), 0])
            pixels['chr2'] = list(bins.iloc[list(pixels['bin2_id']), 0])
            for pair in self.chrom_pair:
                chr1 = list(bins[bins['chrom'] == pair[0]].index)
                chr2 = list(bins[bins['chrom'] == pair[1]].index)
                pair_DS.append(self.__terri_detection_score(pixels, chr1, chr2))
            DS.append(pair_DS)
        df = pd.DataFrame(DS)
        df.columns = [x[0] + '_' + x[1] for x in self.chrom_pair]

        df.to_csv('./file/%s_DS_%s.csv' % (self.describe, str(self.resolution[idx])), sep='\t')
        print('cal random DS')
        # random ds
        random_DS = []
        for cell in clr:

            pair_DS = [0] * 171
            for i in tqdm.tqdm(range(epoch)):
                pixels = cell.pixels()[:]
                for pair in self.chrom_pair:
                    chr1 = list(bins[bins['chrom'] == pair[0]].index)
                    chr2 = list(bins[bins['chrom'] == pair[1]].index)

                    tmp = pixels.sample(len(chr1) + len(chr2), replace=True)
                    pair_DS[self.chrom_pair.index(pair)] += self.__terri_detection_score(tmp, chr1, chr2)
            random_DS.append([x / epoch for x in pair_DS])
        df2 = pd.DataFrame(random_DS)
        df2.columns = [x[0] + '_' + x[1] for x in self.chrom_pair]

        df2.to_csv('./file/%s_terr_rDS_%s.csv' % (self.describe, str(self.resolution[idx])), sep='\t', index=None)

        nDS = df - df2

        nDS.to_csv('./file/%s_terr_nDS_%s.csv' % (self.describe, str(self.resolution[idx])), sep='\t', index=None)
        return nDS

    def __terri_detection_score(self, pixels: pd.DataFrame, chr1: list, chr2: list):
        AA = pixels[pixels['bin1_id'].isin(chr1) & pixels['bin2_id'].isin(chr1)].shape[0]
        BB = pixels[pixels['bin1_id'].isin(chr2) & pixels['bin2_id'].isin(chr2)].shape[0]
        AB = pixels[pixels['bin1_id'].isin(chr1) & pixels['bin2_id'].isin(chr2)].shape[0]
        BA = pixels[pixels['bin1_id'].isin(chr2) & pixels['bin2_id'].isin(chr1)].shape[0]
        Eintra = len(chr1) * len(chr1) + len(chr2) * len(chr2)
        Einter = len(chr1) * len(chr2) * 2
        # print(AA,BB,AB,BA,Eintra,Einter)
        DS = (AA + BB) / Eintra - (AB + BA) / Einter
        return DS

    # 计算每个bin的A/B compartment
    def get_AB_compartment(self, genome='', res=None, epoch=100):
        if self.__merge:
            sys.exit('merge is True, can not get AB_compartmen')
        if res is None:
            sys.exit('res is None, can not get AB_compartmen')
        if res not in self.resolution:
            sys.exit('res is not in resolution, can not get AB_compartmen')

        clr = self.clr_obj[self.resolution.index(res)]

        print('merging cooler')
        merge = []
        for val in self.cache_file:
            name = val + '::resolutions/' + str(res)
            merge.append(name)
        merge_name = './cache/merge_' + str(res) + '_' + self.describe + '.mcool'
        if not os.path.exists(merge_name):
            cooler.merge_coolers(input_uris=merge, output_uri=merge_name, mergebuf=1000000)
        tmp_merge = cooler.Cooler(merge_name)
        cooler.balance_cooler(tmp_merge, store=True)
        print('done')
        genome = bioframe.load_fasta(genome)
        bins = tmp_merge.bins()[:]
        gc_cov = bioframe.frac_gc(bins[['chrom', 'start', 'end']], genome)
        view_df = pd.DataFrame({'chrom': tmp_merge.chromnames,
                                'start': 0,
                                'end': tmp_merge.chromsizes.values,
                                'name': tmp_merge.chromnames}
                               )
        cis_eigs = cooltools.eigs_cis(
            tmp_merge,
            gc_cov,
            view_df=view_df,
            n_eigs=3,
        )

        # cis_eigs[0] returns eigenvalues, here we focus on eigenvectors
        eigenvector_track = cis_eigs[1][['chrom', 'start', 'end', 'E1']]
        store_name = 'E1'
        with tmp_merge.open("r+") as grp:
            if store_name in grp["bins"]:
                del grp["bins"][store_name]
            h5opts = {"compression": "gzip", "compression_opts": 6}
            grp["bins"].create_dataset(store_name, data=eigenvector_track['E1'], **h5opts)

        data = tmp_merge.bins()[:]

        data['E1'] = np.nan_to_num(data['E1'])
        data['sign'] = '$'
        data['sign'][data['E1'] > 0] = 'A'
        data['sign'][data['E1'] < 0] = 'B'
        a_list = []
        b_list = []
        for c in set(data['chrom']):
            seq = list(data[data['chrom'] == c]['sign'])
            base = data[data['chrom'] == c].index[0]

            seq = ''.join(seq)
            #     print(seq)
            new_seq = ''
            idx = []
            tmp_idx = []
            for i in range(len(seq)):
                sym = seq[i]
                if new_seq == '':
                    new_seq += sym
                    tmp_idx.append(i + base)
                else:
                    if sym == seq[i - 1]:
                        tmp_idx.append(i + base)
                    else:
                        idx.append(tmp_idx)
                        tmp_idx = [i + base]
                        new_seq += sym
            for i in range(len(new_seq) - 3):
                if new_seq[i:i + 3] == 'ABA':
                    tmp_A = idx[i] + idx[i + 2]
                    tmp_B = idx[i + 1]
                    a_list.append(tmp_A)
                    b_list.append(tmp_B)

                elif new_seq[i:i + 3] == 'BAB':
                    tmp_A = idx[i + 1]
                    tmp_B = idx[i] + idx[i + 2]
                    a_list.append(tmp_A)
                    b_list.append(tmp_B)
        result = []
        for i in tqdm.tqdm(range(len(clr))):
            cr = clr[i]
            bins = cr.bins()[:]
            pixels = cr.pixels()[:]
            pixels['chr1'] = list(bins.iloc[list(pixels['bin1_id']), 0])
            pixels['chr2'] = list(bins.iloc[list(pixels['bin2_id']), 0])
            pixels['start'] = list(bins.iloc[list(pixels['bin1_id']), 1])
            pixels['end'] = list(bins.iloc[list(pixels['bin1_id']), 2])
            pixels = pixels[pixels['chr1'] == pixels['chr2']]
            DS = []
            for i, j in zip(a_list, b_list):
                DS.append(self.__terri_detection_score(pixels, i, j))
            rDS = self.__return_AB_rds(a_list, b_list, pixels, bins, epoch)
            result.append((DS - rDS).mean())

        return result

    def __return_AB_rds(self, a_list, b_list, pixels, bins, epoch):

        pixels['chr1'] = list(bins.iloc[list(pixels['bin1_id']), 0])
        pixels['chr2'] = list(bins.iloc[list(pixels['bin2_id']), 0])
        pixels['start'] = list(bins.iloc[list(pixels['bin1_id']), 1])
        pixels['end'] = list(bins.iloc[list(pixels['bin1_id']), 2])
        pixels = pixels[pixels['chr1'] == pixels['chr2']]
        bins['index'] = list(bins.index)
        bins = bins.groupby('chrom').describe()['index'].loc[:, ['min', 'max']].astype(int)
        res = []
        for _ in range(epoch):
            for i, j in zip(a_list, b_list):
                idx = np.random.randint(bins['min'], bins['max'])
                idx = random.sample(list(idx), 1)[0]
                tmp = list(sorted(i + j))

                if tmp.index(j[0]) != 0:
                    ch1 = list(range(idx, idx + tmp.index(j[0]))) + list(
                        range(idx + tmp.index(j[-1]) + 1, idx + len(tmp)))
                    ch2 = list(range(idx + tmp.index(j[0]), idx + tmp.index(j[-1]) + 1))
                else:
                    ch1 = list(range(idx, idx + tmp.index(j[-1]) + 1))
                    ch2 = list(range(idx + tmp.index(j[-1]) + 1, idx + len(tmp)))
                try:
                    rds = self.__terri_detection_score(pixels, ch1, ch2)
                except:
                    continue
                res.append(rds)
        return np.array(res).mean()

    def get_TAD_nDS(self, res=500000, epoch=500, nbins=10, ):
        chrom = ['chr' + str(i) for i in range(20)]
        if res not in self.resolution:
            sys.exit('resolution not in %s' % str(self.resolution))

        merge = []
        for val in self.cache_file:
            name = val + '::resolutions/' + str(res)
            merge.append(name)
        merge_name = './cache/merge_' + str(res) + '_' + self.describe + '.mcool'
        if not os.path.exists(merge_name):
            cooler.merge_coolers(input_uris=merge, output_uri=merge_name, mergebuf=1000000)
        tmp_merge = cooler.Cooler(merge_name)
        cooler.balance_cooler(tmp_merge, store=True)
        resolution = res
        win = [10 * resolution]
        insulation_table = insulation(tmp_merge, win, verbose=False)
        boundary = insulation_table[insulation_table['is_boundary_' + str(win[0])] == True]
        
        boundary = boundary[boundary['chrom'].isin(chrom)]

        tad_index = list(boundary.index)

        result = []
        clr = self.clr_obj[self.resolution.index(res)]
        for i in tqdm.tqdm(range(len(clr)), desc='cell'):
            DS = self.__get_TAD_DS(clr[i], tad_index, epoch, nbins)
            result.append(DS)

        return result

    def __get_TAD_DS(self, clr, data, epoch, nbins):
        chrom = ['chr' + str(i) for i in range(20)]
        bins = clr.bins()[:]
        
        bins['index'] = list(bins.index)
        

        thre = bins.groupby('chrom').describe()['index'].loc[:, ['min', 'max']]#.astype(int)
        thre = thre[thre.index.isin(chrom)]
        thre['min'] = thre['min'].astype(int)
        thre['max'] = thre['max'].astype(int)
   
        pixels = clr.pixels()[:]
        pixels['chr1'] = list(bins.iloc[list(pixels['bin1_id']), 0])
        pixels['chr2'] = list(bins.iloc[list(pixels['bin2_id']), 0])
        pixels = pixels[pixels['chr1'] == pixels['chr2']]
        pixels = pixels[pixels['chr1'].isin(chrom)]
        pixels = pixels[pixels['chr2'].isin(chrom)]

        result = pd.DataFrame(columns=['tad_boundary', 'tad_left', 'tad_right', 'chrom'])
        result['tad_boundary'] = list(data)
        result['chrom'] = list(bins.loc[data, :]['chrom'])
        result = result.merge(thre, on='chrom', how='left')
        result['tad_left'] = [max(a, b - nbins) for a, b in zip(list(result['min']), data)]
        result['tad_right'] = [min(a, b + nbins) for a, b in zip(list(result['max']), data)]
        intra_pixels = pixels[pixels['bin1_id'] != pixels['bin2_id']]
        inter_pixels = pixels[pixels['bin1_id'] != pixels['bin2_id']]

        result['DS'] = result.apply(lambda x: self.__get_tad_ds(x, intra_pixels, inter_pixels, nbins), axis=1)

        thre['r_min'] = list(thre['min'] + nbins)
        thre['r_max'] = list(thre['max'] - nbins)

        rds = self.__get_tad_rds(thre, intra_pixels, inter_pixels, nbins, epoch)

        result = np.mean(result['DS']) - rds

        return result

    def __get_tad_rds(self, df, intra_pixels, inter_pixels, nbins, epoch):
        
        res = []
        for i in range(epoch):
            # print(nbins)
            idx = np.random.randint(df['r_min'], df['r_max'])
            idx = random.sample(list(idx), 1)[0]
            idx1 = range(idx - nbins, idx)
            idx2 = range(idx + 1, idx + nbins + 1)
            intra = intra_pixels[((intra_pixels['bin1_id'].isin(idx1)) & (intra_pixels['bin2_id'].isin(idx1))) | (
                        (intra_pixels['bin1_id'].isin(idx2)) & (intra_pixels['bin2_id'].isin(idx2)))]['count'].sum()
            inter = inter_pixels[(inter_pixels['bin1_id'].isin(idx1)) & (inter_pixels['bin2_id'].isin(idx2)) | (
                inter_pixels['bin1_id'].isin(idx2)) & (inter_pixels['bin2_id'].isin(idx1))]['count'].sum()
            # print(idx1,idx2)
            res.append((intra - inter) / (nbins * nbins * 2))

        return np.mean(res)

    def __get_tad_ds(self, x, intra_pixels, inter_pixels, nbins):
        x = list(x)

        idx1 = range(x[1], x[0])
        idx2 = range(x[0] + 1, x[2] + 1)

        intra = intra_pixels[((intra_pixels['bin1_id'].isin(idx1)) & (intra_pixels['bin2_id'].isin(idx1))) | (
                    (intra_pixels['bin1_id'].isin(idx2)) & (intra_pixels['bin2_id'].isin(idx2)))]['count'].sum()
        inter = inter_pixels[(inter_pixels['bin1_id'].isin(idx1)) & (inter_pixels['bin2_id'].isin(idx2)) | (
            inter_pixels['bin1_id'].isin(idx2)) & (inter_pixels['bin2_id'].isin(idx1))]['count'].sum()
        # print(intra, inter, 1)
        return (intra - inter) / (nbins * nbins * 2)

    def get_Centromere_nDs(self, res=None, epoch=100):
        if res is None:
            sys.exit('resoluiton is required')

        clr = self.clr_obj[self.resolution.index(res)]

        result = []
        for i, cr in enumerate(clr):
            log = self.__get_Centromere_nDs(cr, res, epoch=epoch)
            result.append([log[0], log[1], log[2]])
        result = pd.DataFrame(result)
        result.columns = ['DS', 'rDS', 'nDS']
        result.to_csv('./file/%s_centromere_DS_%s_%s.csv' % (self.describe, str(res), self.mcool_names[i]), sep='\t')

        return list(result['nDS'])

    def __get_Centromere_nDs(self, clr, res, epoch):
        centro_pos1 = 3000000  # 3mb
        centro_pos2 = 13000000  # 13mb

        # 获得binid的最大值和最小值，每个染色体不一定有这么长的bp区间
        bin_idx1 = int(centro_pos1 / res)
        bins_idx2 = int(centro_pos2 / res)

        bins = clr.bins()[:]
        pixels = clr.pixels()[:]

        chrom, n_bins = list(bins.groupby('chrom').count()['start'].index), list(bins.groupby('chrom').count()['start'])
        interval_to_cal = {}
        for c, n in zip(chrom, n_bins):
            if bin_idx1 > int(n):
                continue
            interval_to_cal[c] = [bin_idx1, min(n, bins_idx2)]

        ac = 0
        for i, k in enumerate(interval_to_cal.keys()):
            if i == 0:
                ac += n_bins[i]
            else:
                id1, id2 = interval_to_cal[k]
                id1 += ac
                id2 += ac
                ac += n_bins[i]
                interval_to_cal[k] = [id1, id2]
        interval_to_cal_ = []
        for k in interval_to_cal.keys():
            interval_to_cal_ += list(range(interval_to_cal[k][0], interval_to_cal[k][1] + 1))

        chr1 = interval_to_cal_
        chr2 = list(set(list(bins.index)).difference(set(chr1)))

        DS = self.__terri_detection_score(pixels, chr1, chr2)

        r_DS = []
        for i in range(epoch):
            random_interval = list(bins.sample(len(chr1)).index, replace=True)
            random_interval2 = list(set(list(bins.index)).difference(set(random_interval)))
            r_DS.append(self.__terri_detection_score(pixels, random_interval, random_interval2))
        rDS = np.mean(r_DS)
        print(DS - rDS)

        return DS, rDS, DS - rDS

    def get_insulation_score(self, res=None, win=None):
        if res is None:
            sys.exit('resoluiton is required')

        clr = self.clr_obj[self.resolution.index(res)]
        result = []
        for i, cr in enumerate(clr):
            score = self.__get_insulation_score(cr, win)

        return result

    def __get_insulation_score(self, clr, win):
        win = [win]
        if not cooltools.lib.is_cooler_balanced(clr):
            cooler.balance_cooler(clr, store=True)
        # 获得 insulation score，以及是不是边界
        insulation_table = insulation(clr, win, verbose=True)
        s = insulation_table['log2_insulation_score_{}'.format(win[0])]
        s = np.nan_to_num(s)
        return s

    def pooled_cell(self, pool_size=None, num=None, sort=True):
        if self.__merge:
            sys.exit('merged can not pool the cell')
        if self.__pool:
            print('already pooled')
            return

        if pool_size is None or num is None:
            sys.exit('pool_size and num is required')

        # 一共多少个细胞
        element = len(self.clr_obj[0])
        #
        # if sort and num > element:
        #     print('sorted mode: num is bigger than element, set num = element')
        #     num = element
        new_clr_obj = []
        # 循环添加不同分辨率的空list
        for i in range(len(self.resolution)):
            new_clr_obj.append([])

        if not sort:
            for i in tqdm.tqdm(range(num), desc='pooling'):
                pool = random.choices(list(range(element)), k=pool_size)
                mcool_pool = [self.cache_file[idx] for idx in pool]
                for j in range(len(self.resolution)):
                    clr_list = [val + '::/resolutions/' + str(self.resolution[j]) for val in mcool_pool]

                    new_clr_obj[j].append(self.__merge_pool(clr_list, str(i)))
        else:
            # 从大到小排序total contact
            idx = [i[0] for i in sorted(enumerate(self.get_total_contact()), key=lambda x: x[1], reverse=True)]
            # 如果num大于element，那么就随机插入
            if num > element:
                # 假设这是原始列表
                original_list = self.get_total_contact()
                selected_elements = original_list.copy()
                selected_elements.extend(random.sample(original_list, 5))
                sorted_elements = sorted(selected_elements, reverse=True)
                idx = [original_list.index(x) for x in sorted_elements]

            for i in tqdm.tqdm(range(num), desc='pooling::sorted'):
                pool = idx[:i + 1]

                mcool_pool = [self.cache_file[idx] for idx in pool]
                for j in range(len(self.resolution)):
                    clr_list = [val + '::/resolutions/' + str(self.resolution[j]) for val in mcool_pool]
                    new_clr_obj[j].append(self.__merge_pool(clr_list, str(i)))

        self.clr_obj = new_clr_obj
        self.mcool_names = ['pooled_cell_%s' % str(i) for i in range(num)]

    def __merge_pool(self, clr_list, cell_ids):

        cooler.merge_coolers(input_uris=clr_list,
                             output_uri='./cache/%s_pool_all_res-%s.mcool' % (self.describe, cell_ids),
                             mergebuf=1000000)

        res = cooler.Cooler('./cache/%s_pool_all_res-%s.mcool' % (self.describe, cell_ids))

        return res

    def filter_by_total_contact(self, max_=np.inf, min_=0):
        if max_ is None and min_ is None:
            sys.exit('max and min is required')

        tmp_mcool_files = []
        tmp_mcool_names = []
        tmp_cache_file = []
        tmp_clr_obj = []
        for i in range(len(self.resolution)):
            tmp_clr_obj.append([])
        total_contact_list = self.get_total_contact()
        for a, b, c, in zip(self.mcool_files, self.mcool_names, self.cache_file):
            # print(c)
            total_contact = total_contact_list[self.mcool_files.index(a)]
            if total_contact > max_ or total_contact < min_:
                continue
            tmp_mcool_files.append(a)
            tmp_mcool_names.append(b)
            tmp_cache_file.append(c)
            for i in range(len(self.resolution)):
                tmp_clr_obj[i].append(self.clr_obj[i][self.mcool_files.index(a)])

        self.mcool_files = tmp_mcool_files
        self.mcool_names = tmp_mcool_names
        self.cache_file = tmp_cache_file
        self.clr_obj = tmp_clr_obj

    def get_cis_trans_ratio(self, distance=20000):
        if self.resolution[0] > distance:
            sys.exit('distance must be smaller than the samllest resolution')

        result = []
        gap = distance / self.resolution[0]

        for clr in self.clr_obj[0]:
            cis, trans = self.__get_cis_trans_ratio(clr, gap)
            result.append([cis, trans])

        result = pd.DataFrame(result)
        result.columns = ['cis', 'trans']
        result.index = self.mcool_names
        return result

    def __get_cis_trans_ratio(self, clr, distance):
        bins = clr.bins()[:]
        pixels = clr.pixels()[:]
        pixels['chr1'] = list(bins.iloc[list(pixels['bin1_id']), 0])
        pixels['chr2'] = list(bins.iloc[list(pixels['bin2_id']), 0])
        pixels = pixels[pixels['chr1'] == pixels['chr2']]

        cis = sum(pixels[pixels['bin2_id'] - pixels['bin1_id'] <= distance]['count'])
        trans = sum(pixels[pixels['bin2_id'] - pixels['bin1_id'] > distance]['count'])
        return cis, trans

    def sorted_cell(self, reverse=True, num=None):
        idx = [i[0] for i in sorted(enumerate(self.get_total_contact()), key=lambda x: x[1], reverse=reverse)]

        self.cache_file = [self.cache_file[i] for i in idx]
        self.mcool_files = [self.mcool_files[i] for i in idx]
        self.mcool_names = [self.mcool_names[i] for i in idx]
        for i in range(len(self.resolution)):
            self.clr_obj[i] = [self.clr_obj[i][j] for j in idx]

        if num is not None and num < len(self.cache_file):
            self.cache_file = self.cache_file[:num]
            self.mcool_files = self.mcool_files[:num]
            self.mcool_names = self.mcool_names[:num]
            for i in range(len(self.resolution)):
                self.clr_obj[i] = self.clr_obj[i][:num]



   

































